/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#if !defined(__DRIVER_H__)
#define __DRIVER_H__

/**
  * \file common/driver.h
  *
  * \brief This file contains the definition of the interface between the
  * server daemon and the person providing support for a particular type
  * of character LCD display.
  */


typedef void (* pfnDrvCBFanSpeed)(void * context,
                                  unsigned int fan_id,
                                  unsigned int rpm);

typedef void (* pfnDrvCBTemp)(void * context,
                              unsigned int sensor_id,
                              double degrees_c);

typedef void (* pfnDrvCBKeyEvent)(void * context,
                                  unsigned int key_event);

/**
  * \file driver.h
  * 
  * \brief Provides the main interface to which a driver plugin should conform.
  *
  */

#define IILC_DRIVER_API 

/** \brief A pointer to instance data held by the driver implementation.
  *
  * This is really just a better way of writing "void *", as this way the
  * compiler gets to do a bit more type-checking.
  */
typedef struct _tag_driver_instance_data * LP_DRIVER_INSTANCE;

typedef struct _tag_driver_function_array * LP_DRIVER_FN_ARRAY;

/**
  * \brief Nifty little macro to typedef a function pointer which complies with
  * the standards for a driver function pointer. Note that to get the Doxygen
  * generation to be correct, you need to have the config file setup so that
  * ENABLE_PREPROCESSING and MACRO_EXPANSION options are set to 'yes'.
  */
#define DECLARE_DRIVER_INTERFACE_FUNC(name, retVal, args) \
    IILC_DRIVER_API retVal driver_ ## name args ; \
    typedef retVal (* pfnDriver ## name ) args


/**
  * \brief Performs the initial creation of the driver instance.
  *
  * This will be the first function called for a given instance. This function
  * will be called once and once only per device.
  *
  * \param pszPort A string representation of the port which is sufficient to
  * uniquely identify a physical device which this driver can control. Storage
  * is owned by the caller.
  *
  * \param ppDriverInstData pointer to an LP_DRIVER_INSTANCE field which can
  * store context data for the driver module. Storage of the structure pointed
  * to by the return value should be freed during the driver_Dispose() call.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(Create, int, (
                                const char * pszPort,
                                LP_DRIVER_FN_ARRAY pFnArray,
                                LP_DRIVER_INSTANCE * ppDriverInstData
                                ));

/**
  * \brief Performs the final shutdown and cleanup of the driver instance.
  *
  * This function should cleanup any resources it allocated during the
  * driver_Create() function call, or at any point during its lifecycle.
  *
  * \param ppDriverInstData Pointer to an LP_DRIVER_INSTANCE field which
  * has been initialised by a call to the driver_Create() function. It is
  * typically expected that this function sets the contents of this parameter
  * to be NULL upon successful cleanup.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(Dispose, int, (
                                LP_DRIVER_INSTANCE * ppDriverInstData
                                ));


/** \brief Signals to the implementation to commence communications with the
  * physical device.
  *
  * This function will be called after the driver_Create() function, but before
  * any other functions. There are no restrictions against the driver establishing
  * communications before this function call, but it is an error for the driver
  * to produce callbacks until after this function has exited.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(Start, int, (
                                LP_DRIVER_INSTANCE pDriverInstData
                                ));

/** \brief Signals to the implementation to cease communications with the
  * physical device.
  *
  * This function will be called before the driver_Dispose() function. There
  * are no restrictions against the driver maintaining communications after this
  * function call, but it is an error for the driver to produce callbacks after
  * this function has exited.
  *
  * It is permissible for there to be multiple driver_Start(), driver_Stop()
  * cycles during the lifetime (driver_Create() to driver_Dispose()) of a driver
  * instance. The implementation should cope with this eventuality.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(Stop, int, (
                                LP_DRIVER_INSTANCE pDriverInstData
                                ));

/**
  * \brief Instructs the display to set the backlight to the specified
  * brightness level.
  *
  * The brightness level is expressed as a percentage - 0 represents off,
  * 100 represents on. A value outside of this range is an error.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \param nBacklightValue A percentage value, where 0 &lt;= nBacklightValue &lt;= 100
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(SetBacklight, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int nBacklightValue
                                ));

/**
  * \brief Instructs the display to set the contrast to the specified ratio.
  *
  * The contrast ratio is expressed as a percentage - 0 represents off,
  * 100 represents on. A value outside of this range is an error.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \param nContrastValue A percentage value, where 0 &lt;= nContrastValue &lt;= 100
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(SetContrast, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int nBacklightValue
                                ));

/**
  * \brief Sets the specified LED to the given colour.
  *
  * Colour is expressed as a percentage of the three primary additive colours.
  * Each of these values must be in the range 0 to 100 inclusive. An
  * implementation should ignore a colour component if the LED device cannot
  * render that specific colour.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \param led_id ID of the LED to change in the device. This is zero-based.
  *
  * \param nRed 0 &lt;= nRed &lt;= 100, where 0 is off and 100 is fully on
  *
  * \param nGreen 0 &lt;= nGreen &lt;= 100, where 0 is off and 100 is fully on
  *
  * \param nBlue 0 &lt;= nBlue &lt;= 100, where 0 is off and 100 is fully on
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(SetLED, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int led_id,
                                unsigned int nRed,
                                unsigned int nGreen,
                                unsigned int nBlue
                                ));

/**
  * \brief Restarts the physical device, resetting it to the initial power-on
  * status.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(Restart, int, (
                                LP_DRIVER_INSTANCE pDriverInstData
                                ));

/**
  * \brief Clears the screen of the display.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(ClearScreen, int, (
                                LP_DRIVER_INSTANCE pDriverInstData
                                ));

/**
  * \brief Writes the supplied text to the device at the location given.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \param x Horizontal component of the coordinate to write the text. This 
  * metric is measured in character positions. This is zero-based.
  *
  * \param y Vertical component of the coordinate where to write the text. This
  * metric is measured in character positions. This is zero-based.
  *
  * \param strData A string containing the character data to write. Some displays
  * accept the character '\0' as a valid value, which is why this is not a
  * null-terminated string.
  * 
  * \param strLen Length of the string in characters.
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(WriteText, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int x, 
                                unsigned int y, 
                                char * strData,
                                unsigned int strLen
                                ));

/**
  * \brief Sets the specified fan to a given power level.
  *
  * Fan power is expressed as a percentage and the value supplied must be in
  * the range 0 to 100 inclusive.
  *
  * \param pDriverInstData Driver instance data as supplied by a call to the 
  * driver_Create() function.
  *
  * \param fan_id ID of the fan to adjust power output in the device. This is
  * zero-based.
  *
  * \param nPower 0 &lt;= nPower &lt;= 100, where 0 is off and 100 is fully on
  *
  * \retval int zero on success, non-zero otherwise.
  */
DECLARE_DRIVER_INTERFACE_FUNC(SetFanPower, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int fan_id,
                                unsigned int nPower
                                ));


DECLARE_DRIVER_INTERFACE_FUNC(GetHardwareVersion, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int * pnVersionMajor,
                                unsigned int * pnVersionMinor
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(GetScreenSize, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int * pnScreenWidth,
                                unsigned int * pnScreenHeight
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(EnableFanReport, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int fan_id
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(DisableFanReport, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int fan_id
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(EnableTempReport, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int sensor_id
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(DisableTempReport, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                unsigned int sensor_id
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(SetCallbackFan, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                pfnDrvCBFanSpeed pfnDrvCBFanSpeed,
                                void * context
                                ));

DECLARE_DRIVER_INTERFACE_FUNC(SetCallbackTemp, int, (
                                LP_DRIVER_INSTANCE pDriverInstData,
                                pfnDrvCBTemp pfnDrvCBTemp,
                                void * context
                                ));

/*
IILC_DRIVER_API int deviceSetFanReporting(LPDEVICE pdevice, unsigned int mask);

IILC_DRIVER_API int deviceSetTemperatureReporting(LPDEVICE pdevice, unsigned int mask);



IILC_DRIVER_API int deviceSetFanFailsafeTimeout(LPDEVICE pDevice, unsigned int mask, int milliseconds);
*/

typedef struct _tag_driver_function_array
{
    /** Function used to release any resources held by this driver module */
    pfnDriverDispose Dispose;

    /** Function to indicate the driver module should begin comms with the device */
    pfnDriverStart Start;

    /** Function to indicate the driver module should cease comms with the device */
    pfnDriverStop Stop;
        
    pfnDriverSetBacklight SetBacklight;

    pfnDriverWriteText WriteText;

    pfnDriverSetContrast SetContrast;

    pfnDriverSetLED SetLED;

    pfnDriverRestart Restart;

    pfnDriverClearScreen ClearScreen;

    pfnDriverSetFanPower SetFanPower;

    pfnDriverGetHardwareVersion GetHardwareVersion;

    pfnDriverGetScreenSize GetScreenSize;

    pfnDriverEnableFanReport EnableFanReport;

    pfnDriverDisableFanReport DisableFanReport;

    pfnDriverSetCallbackFan SetCallbackFan;

    pfnDriverEnableTempReport EnableTempReport;

    pfnDriverDisableTempReport DisableTempReport;

    pfnDriverSetCallbackTemp SetCallbackTemp;

} DRIVER_FN_ARRAY;




#endif /* __DRIVER_H__ */
