/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#if !defined(__MODULE_H__)
#define __MODULE_H__

/**
  * \file module.h
  *
  * \brief Provides the interface to which a plugin module writer
  * should conform.
  *
  * Plugins using this interface are guaranteed that no two functions 
  * defined within this file will be called concurrently for the same
  * instance. This should avoid the requirement for non-global data to
  * hold locks or require synchronisation primitives to prevent
  * contention between instance-level data.
  *
  * If a module has been loaded multiple times as wishes to share data
  * across instances, however, then shared data must be protected in
  * the usual manner.
  */

#include "lcd.h"

typedef struct _tag_module * LPMODULE;


/*
 * Typedefs for module callbacks
 */

/*
typedef int (* pfnModuleInit)(LPSCREEN pScreen,
                              void ** context);

typedef int (* pfnModuleReadText)(LPSCREEN pScreen,
                                  char ** pWindow);

typedef int (* pfnModuleReadValue)(LPSCREEN pScreen,
                                   unsigned int * nPercentage);

typedef int (* pfnModuleClose)(void * context);
*/




typedef struct _tag_module_func_array * LPMODULE_FN_ARRAY;

typedef int (* pfnModuleCreate)(int moduleHandle,
                                void ** context,
                                LPMODULE_FN_ARRAY pFnArray);


typedef int (* pfnModuleDestroy)(void * context);

/*
 * Callback used to report device temperatures.
 *
 * \param context Context data supplied by the client module during creation.
 * \param sensor_id A zero-based sensor for which the report is being made.
 * \param degrees_c The temperature reported by the sensor.
 */

typedef void (* pfnTemperatureReport)(void * context,
                                      unsigned int sensor_id,
                                      double degrees_c,
                                      const char * pszName);

/*
 * Callback used to query LED status.
 *
 * \param context Context data supplied by the client module during creation.
 * \param led_id Zero-based index of LED to use
 * \param pColourOn Pointer to a colour structure to indicate the ON colour.
 * \param pColourOff Pointer to a colour structure to indicate the OFF colour.
 * \param dutyCycle Percentage value (0-100) to indicate how long the LED is ON for.
 * \param freq Frequency of cycle in Hz/100
 */

/*
 * Callback used to report fans speeds.
 *
 * \param context Context data supplied by the client module during creation.
 * \param fan_number A zero-based index of the fan speed being reported.
 * \param fan_speed The speed of the fan in RPM.
 *
 * \retval An integer - zero on success. Any other value indicates failure.
 */
typedef void (* pfnFanSpeedReport)(void * context,
                                   unsigned int fan_id,
                                   unsigned int rpm,
                                   const char * pszName);


/**
  * This function is called once during the initialisation stage of the module
  * to obtain a list of pointers to the various buffers required to be rendered
  * by the display.
  *
  * \param context The context data which this module supplied during the call to
  *                Create().
  * \param pScreen The screen object which contains information about the display
  *                the text will be rendered to.
  * \param buf     Pointer to a list of (char *) which should be populated by the
  *                function to contain 
  */
typedef int (* pfnGetTextData)(void * context,
                               int nRow,
                               int nWidth,
                               char * buf);

/*
typedef int (* pfnGetNumericData)(void * context,
                                  LPSCREEN pScreen,
                                  unsigned int * value);
*/

typedef struct _tag_module_func_array
{
    /* The function array version this module understands */
    int              version;

    /* Device temperature report callback */
    pfnTemperatureReport pfnTempReport;

    /* Fan speed report callback */
    pfnFanSpeedReport pfnFanSpeedReport;

    /** Used by the server to read data to the screen */
    pfnGetTextData pfnTextData;

    /** Used to obtain the status of each LED this module controls */
    void * /* pfnGetLEDStatus */ pfnReadLEDStatus;

    /** Used to obtain the power required to serve each fan */
    void * /* pfnGetFanPower */ pfnReadFanPower;

} MODULE_FN_ARRAY;



typedef struct _tag_module
{
    /** Handle of the module in memory */
    void *            libraryHandle;

    /** Data supplied by the module used for context */
    void *            moduleContext;

    /** Performs the initial creation of the module */
    pfnModuleCreate   pfnCreate;

    /** Performs the shutdown of the module */
    pfnModuleDestroy  pfnDestroy;

    /* All other functions */
    LPMODULE_FN_ARRAY pFnArray;

    /** Mutex used to serialise communications with the module */
    pthread_mutex_t   mtxModule;

} MODULE;

int moduleInit(const char * moduleName, 
               LPMODULE * ppModule);

/*
LCD_API void moduleTemperatureReport(void * context,
                                    unsigned int sensor_id,
                                    double degrees_c);

LCD_API void moduleFanSpeedReport(void * context,
                                 unsigned int fan_id,
                                 const char * pszFanName,
                                 unsigned int rpm);
*/

/*
LCD_API int moduleGetDisplayData(LPMODULE pModule,
                                 LPSCREEN pScreen,
                                 unsigned int row,
                                 char ** buf);
*/

LCD_API int moduleGetLEDStatus(LPMODULE pModule,
                               unsigned int led_id,
                               LPCOLOUR pColourOn, 
                               LPCOLOUR pColourOff,
                               unsigned int * dutyCycle, 
                               unsigned int * freq);

LCD_API int moduleGetFanPower(LPMODULE pModule,
                              int fan_id,
                              unsigned int * pFanPower);

LCD_API int moduleRender(LPMODULE pModule,
                         int nRow,
                         int nWidth,
                         char * pBuffer);


#endif /* __MODULE_H__ */

