/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */


#include <string.h>

#include "vector.h"
#include "trace.h"


#define IILC_MODULE MODULE_TYPE_APP

#define INITIAL_VECTOR_CAPACITY 20

typedef struct _tag_vector
{
    void ** pData;

    int nCapacity;

    int nSize;
    
    int nHiElement;

} VECTOR;


int vectorCreate(LPVECTOR * ppVector)
{
    int rc = 0;
    LPVECTOR pVector = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_vectorCreate);
    traceDebug(IILC_TRACE_CONTEXT, "ppVector %p", ppVector);

    /* Allocate the memory */
    pVector = (LPVECTOR) malloc(sizeof(VECTOR));
    if (pVector == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        traceError(IILC_TRACE_CONTEXT, "Vector memory allocation failed");
        goto end_of_function;
    }

    /* Init the structure */
    pVector->pData      = NULL;
    pVector->nCapacity  = INITIAL_VECTOR_CAPACITY;
    pVector->nSize      = 0;
    pVector->nHiElement = INITIAL_VECTOR_CAPACITY;

    /* Allocate an initial array */
    pVector->pData = (void **) calloc(pVector->nCapacity, sizeof(void *));
    if (pVector->pData == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        traceError(IILC_TRACE_CONTEXT, "Failed to allocate vector array");
        goto end_of_function;
    }

    /* Assign to the user */
    *ppVector = pVector;

end_of_function:

    /* Cleanup as necessary */
    if (rc) {
        if (pVector) {
            if (pVector->pData) { free(pVector->pData); }
            free(pVector);
        }
    }

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

int vectorDestroy(LPVECTOR * ppVector)
{
    LPVECTOR pVector = *ppVector;

    free(pVector->pData);

    free(pVector);

    *ppVector = NULL;

    return 0;
}

int vectorAdd(LPVECTOR pVector, void * data)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_vectorAdd);
    traceDebug(IILC_TRACE_CONTEXT, "pVector %p, data %p", pVector, data);

    if (pVector->nCapacity == pVector->nSize) {

        void ** pNewArray = NULL;

        /* Create a new array */
        pNewArray = (void **) realloc((void *) pVector->pData, pVector->nCapacity * 2);
        if (pNewArray == NULL) {
            rc = ERR_MEM_ALLOC_FAILED;
            traceError(IILC_TRACE_CONTEXT, "Memory allocation failed");
            goto end_of_function;
        }

        /* Clear out new data region */
        memset(pNewArray[pVector->nCapacity], 0, pVector->nCapacity * sizeof(void *));

        /* Update capacity */
        pVector->nCapacity = pVector->nCapacity * 2;

        /* Remember where new data lies */
        pVector->pData = pNewArray;
    }

    /* Now easy to assign new data */
    pVector->pData[pVector->nSize] = data;

    /* Update size */
    pVector->nSize = pVector->nSize + 1;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

int vectorRemove(LPVECTOR pVector, void * data)
{
    return 0;
}

int vectorGetSize(LPVECTOR pVector, int * pnSize)
{
    *pnSize = pVector->nSize;

    return 0;
}

int vectorGetElementAt(LPVECTOR pVector, int nIndex, void ** pData)
{
    int rc = 0;

    if (nIndex < 0 || nIndex > pVector->nSize) {
        rc = -1;
    }
    else {
        *pData = pVector->pData[nIndex];
    }

    return 0;
}

int vectorSetElementAt(LPVECTOR pVector, int nIndex, void * pData)
{
    int rc = 0;
    int nNewCapacity = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_vectorSetElementAt);
    traceDebug(IILC_TRACE_CONTEXT, "pVector %p, nIndex %d, pData %p", pVector, nIndex, pData);

    nNewCapacity = pVector->nCapacity;
    while (nIndex > nNewCapacity) {
        nNewCapacity *= 2;
    }

    /* Resize array as necessary */
    if (nNewCapacity != pVector->nCapacity) {

        void ** pNewArray = NULL;

        /* Create a new array */
        pNewArray = (void **) realloc((void *) pVector->pData, nNewCapacity);
        if (pNewArray == NULL) {
            rc = ERR_MEM_ALLOC_FAILED;
            traceError(IILC_TRACE_CONTEXT, "Memory allocation failed");
            goto end_of_function;
        }

        /* Clear out new data region */
        memset(pNewArray[pVector->nCapacity], 0, pVector->nCapacity * sizeof(void *));

        /* Remember where new data lies */
        pVector->pData = pNewArray;

        /* Update capacity */
        pVector->nCapacity = nNewCapacity;
    }

    /* Now easy to assign new data */
    pVector->pData[nIndex] = pData;

    /* Update size */
    pVector->nSize = pVector->nSize + 1;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


