/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#   include "time_w32.h"
#else
#endif

#include "lcd.h"
#include "trace.h"
#include "packet.h"
#include "driver.h"
#include "cfontz.h"


#define IILC_MODULE MODULE_TYPE_PACKET


/*
 * Strings which the various CF displays report back their
 * types as.
 */

#define CFONTZ_CFA633_STRING "CFA633"
#define CFONTZ_CFA635_STRING "CFA635"



static int cfontz_WritePacket(LP_CFONTZ_DEV_INSTANCE pDevice,
                              LPPACKET pPacket, 
                              LPPACKET * ppReplyPacket);

static int cfontz_UpdateFanReport(LP_CFONTZ_DEV_INSTANCE pInst);

static int cfontz_UpdateTempReport(LP_CFONTZ_DEV_INSTANCE pInst);

static int cfontz_UpdateFanPower(LP_CFONTZ_DEV_INSTANCE pInst);

static int cfontz_UpdateKeyReport(LP_CFONTZ_DEV_INSTANCE pInst);

static int cfontz_ReceiveTempReport(LP_CFONTZ_DEV_INSTANCE pInst, 
                                    LPPACKET pPacket);

static int cfontz_ReceiveFanSpeedReport(LP_CFONTZ_DEV_INSTANCE pInst,
                                        LPPACKET pPacket);

static int cfontz_ReceiveKeyPressReport(LP_CFONTZ_DEV_INSTANCE pInst, 
                                        LPPACKET pPacket);

/*

static int cfontz_SetCallbackKeyEvent(LP_CFONTZ_DEV_INSTANCE pInst, 
                                      pfnKeyEventListener pfn, 
                                      void * context);

*/

static int cfontz_InitDeviceType(LP_CFONTZ_DEV_INSTANCE pInst,
                                 const char * pszVersionString);


static void * RxThread(void * arg);

static int cfontz_detectSCAB(LP_CFONTZ_DEV_INSTANCE pInst);



/****************************************************************************/
/*                                                                          */
/* Functions which constitute the exported API for the driver module.       */
/*                                                                          */
/****************************************************************************/


int driver_Create(const char * pszPort, LP_DRIVER_FN_ARRAY pFnArray, 
                  LP_DRIVER_INSTANCE * ppDriverInstData)
{
    int rc = 0;
    LP_CFONTZ_DEV_INSTANCE pInst = NULL;

    int i = 0;
    char pszVersionInfo[MAX_DATA_LENGTH + 1];
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_driver_Create);
    traceDebug(IILC_TRACE_CONTEXT,
        "pszPort %p, pFnArray %p, ppDevice %p", pszPort, pFnArray, ppDriverInstData);

    /* Sanity check */
    if (pszPort == NULL || pFnArray == NULL || ppDriverInstData == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Input parameter NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT, "Creating CrystalFontz device on port [%s]", pszPort);

    /* Allocate the main memory with simple clear */
    pInst = (LP_CFONTZ_DEV_INSTANCE) calloc(1, sizeof(CFONTZ_DEV_INSTANCE));

    /*
     * More complicated init
     */

    /* Take a copy of the port string */
    pInst->pszPort = (char *) malloc(strlen(pszPort) + 1);
    strcpy(pInst->pszPort, pszPort);

    /*
     * Setup the threads and sync primitives
     */

    /* Must wait to init RxThreadID later */

    /* Mutex to lock when waiting for an Rx packet */
    pthread_mutex_init(&(pInst->RxMutex), NULL);

    /* Semaphore on which to wait when waiting for an Rx packet */
    pthread_cond_init(&(pInst->RxSemaphore), NULL);

    /* Flag to indicate the Rx thread should stay running */
    pInst->RxRunning = 1;

    /* Mutex to lock while the Rx thread is runnning */
    pthread_mutex_init(&(pInst->RunningMutex), NULL);

    /*
     * Callback information
     */

    /* Function pointer block */
    pInst->pfn.FanSpeed = NULL;
    pInst->pfn.KeyEvent = NULL;
    pInst->pfn.Temp     = NULL;

    /* Context info block */
    pInst->context.FanSpeed = NULL;
    pInst->context.KeyEvent = NULL;
    pInst->context.Temp     = NULL;

/***************************************************************/

    /* Update the function pointers */
    pFnArray->Dispose       = driver_Dispose;
    pFnArray->Start         = driver_Start;
    pFnArray->Stop          = driver_Stop;
    pFnArray->SetBacklight  = driver_SetBacklight;
    pFnArray->GetScreenSize = driver_GetScreenSize;
    pFnArray->SetFanPower   = driver_SetFanPower;
    pFnArray->WriteText     = driver_WriteText;
    pFnArray->EnableFanReport = driver_EnableFanReport;
    pFnArray->DisableFanReport = driver_DisableFanReport;
    pFnArray->SetCallbackFan = driver_SetCallbackFan;
    pFnArray->SetContrast    = driver_SetContrast;
    pFnArray->EnableTempReport = driver_EnableTempReport;
    pFnArray->DisableTempReport = driver_DisableTempReport;
    pFnArray->SetCallbackTemp = driver_SetCallbackTemp;

    /*
     * Data structure init complete - now start doing some work to get this
     * unit running
     */

    /* Perform OS-specific init to physical device */
    rc = cfontz_InitDevice(pInst);
    if (rc) {
        goto end_of_function;
    }

    /* Pause required to allow device to initialise */
	SLEEP(350);

    /* Spin off a receiving thread */
    rc = pthread_create(&(pInst->RxThreadID), NULL, RxThread, (void *) pInst);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "pthread_create error %d (%s)\n", errno, strerror(errno));
        goto end_of_function;
    }

    /*
     * Send a command up to the display to get the hardware version
     */
    
    /* Create named packet type with no data */
    packetCreate(PACKET_TYPE_GET_VERSION_INFO, &pPacket);
    cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    packetDispose(&pPacket);

    /* Check the return packet */
    if (pReplyPacket != NULL) {
        if (packetGetClass(pReplyPacket) != PACKET_CLASS_NORMAL_RESPONSE) {
            traceError(IILC_TRACE_CONTEXT, "Invalid response to hardware version request");
        }
    }
    else {
        traceError(IILC_TRACE_CONTEXT, "No response to hardware version request");
    }


    /* Zap data buffer to NULLs */
    memset((void *) pszVersionInfo, 0, MAX_DATA_LENGTH + 1);

    /* Get the hardware string */
    while (packetReadData(pReplyPacket, (unsigned char *) &pszVersionInfo[i++]) != ERR_END_OF_PACKET);

    /* Dump out the hardware version info */
    traceDebug(IILC_TRACE_CONTEXT, "Connected to CrystalFontz hardware [%s]", pszVersionInfo);

    /* Now try to programmatically work out what we're connected to */
    rc = cfontz_InitDeviceType(pInst, pszVersionInfo);
    if (rc) { goto end_of_function; }

    /* Create the fan power array */
    if (pInst->nFanCount > 0) {

        /* All fans default to 100% power */
        pInst->nFanPower = (unsigned int *) calloc(pInst->nFanCount, sizeof(unsigned int));
        for (i = 0; i < pInst->nFanCount; i++) {
            pInst->nFanPower[i] = 100;
        }
    }

    /*
	{
		char * envVar = NULL;
		unsigned int nPower = 100;

		envVar = getenv("FANPOWER");
		if (envVar != NULL) {
			nPower = atoi(envVar);
		}
		else {
			printf("FANPOWER env var not set\n");
		}

    	driver_SetFanPower((LP_DRIVER_INSTANCE) pInst, 0, nPower);
	}
    */

    /* All OK - assign to caller's data area */
    *ppDriverInstData = (LP_DRIVER_INSTANCE) pInst;    

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
	return rc;

}

int cfontz_detectSCAB(LP_CFONTZ_DEV_INSTANCE pInst)
{
    int rc = 0;

    unsigned char gpio = 0;
    unsigned char state = 0;

    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    /* Assume not present */
    pInst->bSCABPresent = 0;

    /* Taken from the "Detecting SCAB" thread post dated 10-04-2005 11:05 PM,
     * found in the "Product questions" forum.
     *
     * Since the SCAB has a 1K pull-up on GPIO[4] (the DOW line) it should be
     * possible to use that to detect if it is connected:
     *
     * 1) set GPIO[4] to drive high, pull low
     * 2) drive it low (meaning it is pulled low by the internal 5K)
     * 3) read it. If it is low, then no SCAB is attached, if it is high
     *    (1K SCAB pull-up swamping internal 5K pull-down) then a SCAB is
     *    attached.
     * 4) restore GPIO[4]'s port state.
     */

    /* Configure GPIO[4] */
    packetCreate(PACKET_TYPE_SET_GPIO, &pPacket);
    packetWriteData(pPacket, 4);
    packetWriteData(pPacket, 0);
    packetWriteData(pPacket, 8);
    cfontz_WritePacket(pInst, pPacket, &pReplyPacket);

    /* Check that worked OK */
    if (packetGetClass(pReplyPacket) != PACKET_CLASS_NORMAL_RESPONSE) {
        /* Abort detection - assume not present */
    }

    /* Dispose */
    packetDispose(&pPacket);
    packetDispose(&pReplyPacket);

    /* Read GPIO[4] */
    packetCreate(PACKET_TYPE_GET_GPIO, &pPacket);
    packetWriteData(pPacket, 4); /* GPIO 4 */
    cfontz_WritePacket(pInst, pPacket, &pReplyPacket);

    /* Check that worked OK */
    if (packetGetClass(pReplyPacket) != PACKET_CLASS_NORMAL_RESPONSE) {
        /* Abort detection - assume not present */
    }

    /* Read data */
    packetReadData(pReplyPacket, &gpio);
    packetReadData(pReplyPacket, &state);

    if (gpio == 4 && state > 0) {
        /* SCAB detected */
        pInst->bSCABPresent = 1;
    }

    /* Dispose */
    packetDispose(&pPacket);
    packetDispose(&pReplyPacket);

    /* Restore GPIO function to normal */
    packetCreate(PACKET_TYPE_SET_GPIO, &pPacket);
    packetWriteData(pPacket, 4);
    packetWriteData(pPacket, 0);
    packetWriteData(pPacket, 7);
    cfontz_WritePacket(pInst, pPacket, &pReplyPacket);

    /* Dispose */
    packetDispose(&pPacket);
    packetDispose(&pReplyPacket);

    return (rc);
}


int driver_Dispose(LP_DRIVER_INSTANCE * ppDriverInstData)
{
    int rc = 0;
    
    LP_CFONTZ_DEV_INSTANCE * ppInst = (LP_CFONTZ_DEV_INSTANCE *) ppDriverInstData;
    LP_CFONTZ_DEV_INSTANCE   pInst  = NULL;
    void * pTermValue = NULL;

    if (ppInst == NULL || *ppInst == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Convenience */
    pInst = *ppInst;

    /* Switch off key reporting */
    pInst->keyPressMask   = 0;
    pInst->keyReleaseMask = 0;
    cfontz_UpdateKeyReport(pInst);

    /* Switch off temperature reporting */
    pInst->tempReportMask = 0;
    cfontz_UpdateTempReport(pInst);

    /* Switch off fan reporting */
    pInst->fanReportMask = 0;
    cfontz_UpdateFanReport(pInst);

    /* Stop the receive thread running */
    pInst->RxRunning = 0;
    
    /* Wait for the thread to terminate */
    pthread_join(pInst->RxThreadID, &pTermValue);

    /* Shutdown the physical device */
#if defined(_WIN32)
    CloseHandle(pInst->Device);
#else
    close(pInst->Device);
#endif

    /* Release all memory */


    if (pInst->pszPort != NULL) {
        free(pInst->pszPort);
    }

    free(pInst);

end_of_function:

    return rc;
}


int driver_Start(LP_DRIVER_INSTANCE pDriverInstData)
{
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    pInst->bStarted = 1;

    return 0;
}


int driver_Stop(LP_DRIVER_INSTANCE pDriverInstData)
{
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    pInst->bStarted = 0;

    return 0;
}


int driver_GetScreenSize(LP_DRIVER_INSTANCE pDriverInstData,
                         unsigned int * pnScreenWidth, unsigned int * pnScreenHeight)
{
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    *pnScreenHeight = pInst->nHeight;
    *pnScreenWidth  = pInst->nWidth;

    return 0;
}


int driver_SetLED(LP_DRIVER_INSTANCE pDriverInstData, unsigned int led_id,
                  unsigned int nRed, unsigned int nGreen, unsigned int nBlue)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displaySetLED);
    traceDebug(IILC_TRACE_CONTEXT,
        "pInst %p, led_id %u, nRed %u, nGreen %u, nBlue %u",
        pInst, led_id, nRed, nGreen, nBlue);

    /* Ignore the blue content of the colour */

    /* Write out the green component */
    packetCreate(PACKET_TYPE_SET_GPIO, &pPacket);
    packetWriteData(pPacket, 11 - (led_id * 2));
    packetWriteData(pPacket, (unsigned char) nGreen);
    cfontz_WritePacket(pInst, pPacket, NULL);
    packetDispose(&pPacket);

    /* Write out the red component */
    packetCreate(PACKET_TYPE_SET_GPIO, &pPacket);
    packetWriteData(pPacket, 12 - (led_id * 2));
    packetWriteData(pPacket, (unsigned char) nRed);
    cfontz_WritePacket(pInst, pPacket, NULL);
    packetDispose(&pPacket);

    IILC_TRACE_EXIT_RC(rc);
    return (rc);
}


int driver_SetFanFailsafeTimeout(LP_DRIVER_INSTANCE pDriverInstData, unsigned int mask, int milliseconds)
{
    int rc = 0, timeout = 0;
    PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    /* Calculate how many 1/8 second are in 'milliseconds' (1/8s == 125ms) */
    timeout = milliseconds / 125;

    /* Ensure we pass a good value in */
    if (timeout == 0)       { timeout++; }
    else if (timeout > 255) { timeout = 255; }

    rc = packetCreate(PACKET_TYPE_SET_FAN_FAILSAFE, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, mask);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, timeout);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_FAN_FAILSAFE | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}




/****************************************************************************/
/*                                                                          */
/* Internal functions used to perform various internal operations specific  */
/* to this particular driver.                                               */
/*                                                                          */
/****************************************************************************/




int cfontz_WritePacket(LP_CFONTZ_DEV_INSTANCE pDevice, LPPACKET pPacket, LPPACKET * ppReplyPacket)
{
    int rc = 0;
    int run, ret;

    struct timespec timeToWait;

    unsigned char buf[1024];
    int bufLen = sizeof(buf) / sizeof(buf[0]);
    int dataLen = bufLen;

    LPPACKET pReplyPacket = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_deviceWritePacket);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDevice %p, pPacket %p, ppReplyPacket %p", pDevice, pPacket, ppReplyPacket);

    IILC_ASSERT(pDevice != NULL);
    IILC_ASSERT(pPacket != NULL);


    /* Flatten the packet into a buffer */
    rc = packetFlatten(pPacket, buf, &dataLen);
    if (rc) {
        traceDebug(IILC_TRACE_CONTEXT, "Failure while flattening packet %d", rc);
        goto end_of_function;
    }


    /* ----- Send it to the device ----- */

	/* Repeat this operation a few times */
	for (run = 0; run < 10; run++) {

#if defined(_WIN32)
        rc = WriteFile(pDevice->Device, buf, dataLen, &ret, NULL);
        if (!rc) {
            printf("WriteFile failed\n");
        }
#else
		ret = write(pDevice->Device, buf, dataLen);
#endif

		if (ret < 0) {
			traceDebug(IILC_TRACE_CONTEXT, "write %d %s", ret, strerror(errno));
		}
		else {
			break;
		}
		if (run > 0) {
			traceDebug(IILC_TRACE_CONTEXT, "write(): EAGAIN #%d", run);
		}

		SLEEP(1);
	}
	
	if (ret < 0) {
		traceDebug(IILC_TRACE_CONTEXT, "write() failed; %s", strerror(errno));
        goto end_of_function;
	}
    else if (ret != dataLen) {
		traceDebug(IILC_TRACE_CONTEXT, "partial write(): len=%d ret=%d", dataLen, ret);
        goto end_of_function;
	}


    /* ----- Now wait for a response packet ----- */


    /* Lock the mutex */
    rc = pthread_mutex_lock(&pDevice->RxMutex);
    if (rc) {
        printf("Rx mutex lock failed: %s\n", strerror(errno));
        goto end_of_function;
    }

    /* Wait for a maximum of 250ms (which is what the PDF user doc says) */
    {
        unsigned int milliseconds = 250;

        clock_gettime(CLOCK_REALTIME, &timeToWait);

        timeToWait.tv_nsec += (milliseconds % 1000) * 1000000;
        timeToWait.tv_sec  += (milliseconds / 1000);

        /* Check within range */
        if (timeToWait.tv_nsec > 1000000000) {
            timeToWait.tv_sec++;
            timeToWait.tv_nsec -= 1000000000;
        }
    }

    /* Wait on the semaphore */
    rc = pthread_cond_timedwait(&(pDevice->RxSemaphore), &(pDevice->RxMutex), &timeToWait);

    if (pDevice->pReplyPacket == NULL) {
        printf("No reply packet! rc = %d\n", rc);
        exit(1);
    }

    if (rc == ETIMEDOUT) {
        /* Create a time-out packet for the user */
        packetCreate(0xFF, &(pDevice->pReplyPacket));
    }
    else if (rc) {
        printf("Wait on semaphore failed: %s\n", strerror(errno));
        goto end_of_function;
    }

    /* Grab the returned packet and clear */
    pReplyPacket = pDevice->pReplyPacket;
    pDevice->pReplyPacket = NULL;

    /* Release the Rx mutex */
    rc = pthread_mutex_unlock(&(pDevice->RxMutex));
    if (rc) {
        printf("Rx mutex unlock failed: %s\n", strerror(errno));
        goto end_of_function;
    }


    /* Return packet to the user? */
    if (pReplyPacket != NULL) {
        /* Does the user wish to know about the returned packet? */
        if (ppReplyPacket != NULL) {
            *ppReplyPacket = pReplyPacket;
        }
        else {
            packetDispose(&pReplyPacket);
        }
    }
	
end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


/*
 * Continuously-running thread which handles comms coming from the device.
 * Dispatches the received packets either back to the calling thread, or
 * to handlers of async events.
 */
void * RxThread(void * arg)
{
    LP_CFONTZ_DEV_INSTANCE pDevice = (LP_CFONTZ_DEV_INSTANCE) arg;

    int bytesRead = 0;
    int bytesAvail = 0;
    int ret;

    unsigned char buf[1024];
    int bufLen = sizeof(buf) / sizeof(buf[0]);

	do {
		int rc = 0;
		LPPACKET pReplyPacket = NULL;

		/* Read from the port */
#if defined(_WIN32)
        rc = ReadFile(pDevice->Device, &buf[bytesAvail], bufLen - bytesAvail, &ret, NULL);
        if (!rc) {
            printf("ReadFile failed\n");
        }
#else
		ret = read(pDevice->Device, &buf[bytesAvail], bufLen - bytesAvail);
#endif

        if (ret < 0 && errno == EAGAIN) {
            /* Snooze for 10ms while we wait for something to turn up */
            SLEEP(10);
            continue;
        }

        /* We now have more in the buffer */
        bytesAvail += ret;

        /* Nothing has been read from the buffer yet */
        bytesRead = 0;

        do {

            /* Attempt to create a packet from bytes avaialble in this buffer */
            rc = packetInflate(buf, bytesAvail, &bytesRead, &pReplyPacket);

            /* Remove any bytes that have been read */
            if (bytesRead > 0) {
                bytesAvail -= bytesRead;
                memmove(buf, buf + bytesRead, bytesAvail);
            }

            /* Have we succeeded? */
            if (rc == 0) {

                /* What sort of packet is this? */
                int packetClass = packetGetClass(pReplyPacket);
                switch (packetClass) {

                /* Async report from device to app */
                case PACKET_CLASS_NORMAL_REPORT:
                {
                    /* Get the packet type */
                    PACKET_TYPE packetType;
                    packetGetType(pReplyPacket, &packetType);

                    /* Ensure we've been started */
                    if (pDevice->bStarted) {

                        /* Decide what to do with this */
                        switch (packetType) {

                        case PACKET_TYPE_RPT_TEMP_SENSOR:
                            cfontz_ReceiveTempReport(pDevice, pReplyPacket);
                            break;

                        case PACKET_TYPE_RPT_FAN_SPEED:
                            cfontz_ReceiveFanSpeedReport(pDevice, pReplyPacket);
                            break;

                        case PACKET_TYPE_RPT_KEY_ACTIVITY:
                            cfontz_ReceiveKeyPressReport(pDevice, pReplyPacket);
                            break;

                        default:
                            printf("Unknown packet type (%X)\n", packetType);
                            break;
                        }
                    }
                    break;
                }

                /* Both normal and error responses to be handed back to waiting caller */
                case PACKET_CLASS_NORMAL_RESPONSE:
                case PACKET_CLASS_ERROR_RESPONSE:
                {
                    PACKET_TYPE ptype;
                    packetGetType(pReplyPacket, &ptype);

                    /* Lock the Rx mutex */
                    rc = pthread_mutex_lock(&pDevice->RxMutex);
                    if (rc) {
                        printf("Rx mutex lock failed: %s\n", strerror(errno));
                        break;
                    }

                    /* Set the reply packet (memory now owned by the device) */
                    pDevice->pReplyPacket = pReplyPacket;
                    pReplyPacket = NULL;

                    /* Signal the Rx semaphore */
                    rc = pthread_cond_signal(&(pDevice->RxSemaphore));
                    if (rc) {
                        printf("Rx semaphore signal failed: %s\n", strerror(errno));
                        break;
                    }

                    /* Release the Rx mutex */
                    rc = pthread_mutex_unlock(&(pDevice->RxMutex));
                    if (rc) {
                        printf("Rx mutex unlock failed: %s\n", strerror(errno));
                        break;
                    }

                    break;
                }

                /* Shouldn't get this packet received! */
                case PACKET_CLASS_NORMAL_REQUEST:
                    printf("Strange packet type received\n");
                    break;

                /* Broken!! */
                default:
                    printf("Unable to determine packet class\n");
                    break;
                }

                /* Dispose of the packet */
                if (pReplyPacket != NULL) {
                    packetDispose(&pReplyPacket);
                }
            }
            else if (rc == ERR_INSUFFICIENT_DATA) {
                /* Just a question of waiting for more? */
                SLEEP(10);
                break; /* out of bytesAvail loop */
            }
            else {
                /* Must be bad */
                printf("Fatal error from packet inflate %d\n", rc);
                break;
            }

        } while (bytesAvail > 0);

    } while (pDevice->RxRunning);
	
	return NULL;
}

int driver_SetBacklight(LP_DRIVER_INSTANCE pDriverInstData, unsigned int backlightValue)
{
    int rc = 0;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    rc = packetCreate(PACKET_TYPE_SET_BACKLIGHT, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, backlightValue);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }
    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}

int driver_SetContrast(LP_DRIVER_INSTANCE pDriverInstData, unsigned int contrastValue)
{
    int      rc      = 0;
    LPPACKET pPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    rc = packetCreate(PACKET_TYPE_SET_LCD_CONTRAST, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, contrastValue);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, NULL);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket != NULL) { packetDispose(&pPacket); }

    return rc;
}


int driver_ClearScreen(LP_DRIVER_INSTANCE pDriverInstData)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LPPACKET pReplyPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    rc = packetCreate(PACKET_TYPE_CLEAR_SCREEN, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket != NULL) { packetDispose(&pPacket); }
    if (pReplyPacket != NULL) { packetDispose(&pReplyPacket); }

    return rc;
}



int driver_WriteText(LP_DRIVER_INSTANCE pDriverInstData, unsigned int x, unsigned int y, char * strData, unsigned int width)
{
    int rc = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    rc = packetCreate(PACKET_TYPE_WRITE_DATA, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, x);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, y);
    if (rc) { goto end_of_function; }

    rc = packetWriteString(pPacket, strData, width);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket)      { packetDispose(&pPacket); }
    if (pReplyPacket) { packetDispose(&pReplyPacket); }

    return rc;
}



int driver_Restart(LP_DRIVER_INSTANCE pDriverInstData)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    rc = packetCreate(PACKET_TYPE_POWER_CONTROL, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, 8);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, 18);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, 99);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, NULL);
    if (rc) { goto end_of_function; }

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    return rc;
}

int driver_SetFanPower(LP_DRIVER_INSTANCE pDriverInstData, unsigned int fan_id, unsigned int nPower)
{
    int rc = 0;

    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    IILC_TRACE_ENTRY(IILC_MODULE, T_driver_SetFanPower);
    traceDebug(IILC_TRACE_CONTEXT, "pInst %p, fan_id %u, nPower %u", pInst, fan_id, nPower);

    /* Sanity check */
    if (pDriverInstData == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Fans numbered from 0 */
    if (fan_id >= pInst->nFanCount) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Should be a percentage */
    if (nPower > 100) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Keep track internally */
    pInst->nFanPower[fan_id] = nPower;

    /* Update the device */
    cfontz_UpdateFanPower(pInst);

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

int cfontz_UpdateFanReport(LP_CFONTZ_DEV_INSTANCE pInst)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LPPACKET pReplyPacket = NULL;

    /* Send to the display */
    rc = packetCreate(PACKET_TYPE_SET_FAN_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, pInst->fanReportMask);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket) packetDispose(&pPacket);
    if (pReplyPacket) packetDispose(&pReplyPacket);

    return rc;
}


int cfontz_UpdateKeyReport(LP_CFONTZ_DEV_INSTANCE pInst)
{
    int rc = 0;
    LPPACKET pPacket = NULL;
    LPPACKET pReplyPacket = NULL;

    /* Send to the display */
    rc = packetCreate(PACKET_TYPE_SET_KEY_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, pInst->keyPressMask);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, pInst->keyReleaseMask);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

end_of_function:

    if (pPacket) packetDispose(&pPacket);
    if (pReplyPacket) packetDispose(&pReplyPacket);

    return rc;
}

int driver_EnableFanReport(LP_DRIVER_INSTANCE pDriverInstData, unsigned int fan_id)
{
    int rc = 0;

    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    /* Setup the mask & update */
    pInst->fanReportMask |= (1 << fan_id);
    rc = cfontz_UpdateFanReport(pInst);

    return rc;
}

int driver_DisableFanReport(LP_DRIVER_INSTANCE pDriverInstData, unsigned int fan_id)
{
    int rc = 0;

    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    /* Setup the mask & update */
    pInst->fanReportMask &= ~(1 << fan_id);
    rc = cfontz_UpdateFanReport(pInst);

    return rc;
}

int driver_EnableTempReport(LP_DRIVER_INSTANCE pDriverInstData, unsigned int sensor_id)
{
    int rc = 0;

    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    /* Setup the mask & update */
    pInst->tempReportMask |= (1 << sensor_id);
    rc = cfontz_UpdateTempReport(pInst);

    return rc;
}

int driver_DisableTempReport(LP_DRIVER_INSTANCE pDriverInstData, unsigned int sensor_id)
{
    int rc = 0;

    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    /* Setup the mask & update */
    pInst->tempReportMask &= ~(1 << sensor_id);
    rc = cfontz_UpdateTempReport(pInst);

    return rc;
}


int driver_SetTemperatureReporting(LP_DRIVER_INSTANCE pDriverInstData, unsigned int mask)
{
    int rc = 0;
    //PACKET_TYPE replyType = 0;
    LPPACKET pPacket = NULL, pReplyPacket = NULL;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    rc = packetCreate(PACKET_TYPE_SET_TEMP_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    rc = packetWriteData(pPacket, mask & 0x000000FF);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0x0000FF00) >> 8);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0x00FF0000) >> 16);
    if (rc) { goto end_of_function; }

    rc = packetWriteData(pPacket, (mask & 0xFF000000) >> 24);
    if (rc) { goto end_of_function; }

    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    /*
    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }
    */

end_of_function:

    /* Clean up resources */
    if (pPacket != NULL) {
        (void) packetDispose(&pPacket);
    }

    if (pReplyPacket != NULL) {
        (void) packetDispose(&pReplyPacket);
    }

    return rc;
}

/**
 * Called when a fan speed report is received from the connected display.
 */
int cfontz_ReceiveFanSpeedReport(LP_CFONTZ_DEV_INSTANCE pInst, LPPACKET pPacket)
{
    int rc = 0;

    /* Final fan speed value */
    unsigned int fanSpeed = 0;
    
    /* Raw data from the packet */
    unsigned char LSB, MSB, fan_number, number_of_fan_tach_cycles;

    /* Pulses Per Revolution  - specific to each fan, most commonly 2 */
    const int PPR = 2;

    /* Join of MSB+LSB */
    unsigned short Fan_Timer_Ticks;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayReceiveFanSpeedReport);
    traceDebug(IILC_TRACE_CONTEXT, "pInst %p, pPacket %p", pInst, pPacket);

    /* Do we have any listeners? */
    if (pInst->pfn.FanSpeed == NULL) {
        goto end_of_function;
    }

    /* Read the raw data from the packet */
    if (!rc) { rc = packetReadData(pPacket, &fan_number); }
    if (!rc) { rc = packetReadData(pPacket, &number_of_fan_tach_cycles); }
    if (!rc) { rc = packetReadData(pPacket, &LSB); }
    if (!rc) { rc = packetReadData(pPacket, &MSB); }
    if (rc) {
        traceWarning(IILC_TRACE_CONTEXT, "Badly-formed fan speed report packet");
        goto end_of_function;
    }

    /* Convert raw data to something useful */
    Fan_Timer_Ticks = (MSB << 8) + LSB;

    /* Useful information dumped our to trace */
    traceDebug(IILC_TRACE_CONTEXT,
        "fan_number " T_FMT_UINT " Fan_Timer_Ticks " T_FMT_UINT " Tach_Cycles " T_FMT_UINT,
        (unsigned int) fan_number, (unsigned int) Fan_Timer_Ticks, number_of_fan_tach_cycles);

    if (number_of_fan_tach_cycles <= 3) {        
        fanSpeed = 0; /* Stopped */
    }
    else if (number_of_fan_tach_cycles <= 4) {        
        fanSpeed = 1; /* Too slow to measure */
    }
    else if (number_of_fan_tach_cycles == 0xFF) {
        fanSpeed = 0XFFFFFFFF; /* Out of range */
    }
    else {
        /* Use the formula provided by CF-635 data sheet v 1.4 */
        fanSpeed = ((27692308L / PPR) * (unsigned long)(number_of_fan_tach_cycles - 3)) / (Fan_Timer_Ticks);
    }

    traceDebug(IILC_TRACE_CONTEXT, "fan speed " T_FMT_UINT, fanSpeed);

    /* Send this data to the callback */
    (pInst->pfn.FanSpeed)(pInst->context.FanSpeed, (unsigned int) fan_number, fanSpeed);

end_of_function:

    IILC_TRACE_EXIT;
    return rc;
}


int cfontz_ReceiveKeyPressReport(LP_CFONTZ_DEV_INSTANCE pInst, LPPACKET pPacket)
{
    unsigned char event;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayReceiveKeyPressReport);
    traceDebug(IILC_TRACE_CONTEXT, "pInst %p, pPacket %p", pInst, pPacket);

    /* Ignore if no-one is bothered about receiving reports */
    if (pInst->pfn.KeyEvent == NULL) {
        goto end_of_function;
    }

    /* Read the data */
    packetReadData(pPacket, &event);

    /* Perform the callback */
    (pInst->pfn.KeyEvent)(pInst->context.KeyEvent, 0);

end_of_function:

    IILC_TRACE_EXIT;
    return 0;
}


/**
 * Called when a temperature report is received from the connected display.
 */
int cfontz_ReceiveTempReport(LP_CFONTZ_DEV_INSTANCE pInst, LPPACKET pPacket)
{
    int rc = 0;
    //int i = 0;
    double degc = 0.0;
    unsigned char sensor, MSB, LSB, DOW_crc_status;
    unsigned short report;

    IILC_TRACE_ENTRY(IILC_MODULE, T_cfontz_ReceiveTempReport);
    traceDebug(IILC_TRACE_CONTEXT, "pInst %p, pPacket %p", pInst, pPacket);

    /* Do we have any callbacks registered? */
    if (pInst->pfn.Temp == NULL) {
        traceDebug(IILC_TRACE_CONTEXT, "No temp callbacks registered");
        goto end_of_function;
    }

    /* Read the raw data */
    packetReadData(pPacket, &sensor);
    packetReadData(pPacket, &LSB);
    packetReadData(pPacket, &MSB);
    packetReadData(pPacket, &DOW_crc_status);

    /* Check CRC is OK */
    if (DOW_crc_status == 0) {
        printf("Bad DOW CRC\n");
        rc = -1;
        goto end_of_function;
    }

    /* Convert to a full 16-bit value and get true reading */
    report = (MSB << 8) + LSB;
    degc = ((double) report) / 16.0; 

    (pInst->pfn.Temp)(pInst->context.Temp, sensor, degc);

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}



int driver_SetCallbackTemp(LP_DRIVER_INSTANCE pDriverInstData, pfnDrvCBTemp pfn, void * context)
{
    int rc = 0;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    pInst->pfn.Temp     = pfn;
    pInst->context.Temp = context;

    return rc;   
}

int driver_SetCallbackFan(LP_DRIVER_INSTANCE pDriverInstData, pfnDrvCBFanSpeed pfn, void * context)
{
    int rc = 0;
    LP_CFONTZ_DEV_INSTANCE pInst = (LP_CFONTZ_DEV_INSTANCE) pDriverInstData;

    pInst->pfn.FanSpeed     = pfn;
    pInst->context.FanSpeed = context;

    return rc;
}

int cfontz_SetCallbackKeyEvent(LP_CFONTZ_DEV_INSTANCE pInst, pfnDrvCBKeyEvent pfn, void * context)
{
    pInst->pfn.KeyEvent     = pfn;
    pInst->context.KeyEvent = context;
    return 0;
}


int cfontz_InitDeviceType(LP_CFONTZ_DEV_INSTANCE pInst, const char * pszVersionString)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_cfontz_InitDeviceType);

    if (strncmp(pszVersionString, CFONTZ_CFA635_STRING, strlen(CFONTZ_CFA635_STRING)) == 0) {

        /* CFA-635 */
        pInst->deviceType = CFONTZ_DEVICE_TYPE_CFA635;
        pInst->nHeight   = 4;
        pInst->nWidth    = 20;
        pInst->nLEDCount = 4;
        pInst->nFanCount = 0;

        /* Do we have a SCAB present? */
        rc = cfontz_detectSCAB(pInst);
        if (rc) { goto end_of_function; }

        /* Some init can only happen with a SCAB present */
        if (pInst->bSCABPresent) {

            /* No way of really knowing for sure how many fans are attached - max 4 though */
            pInst->nFanCount = 4; 
        }
    }
    else if (strncmp(pszVersionString, CFONTZ_CFA633_STRING, strlen(CFONTZ_CFA633_STRING)) == 0) {

        /* CFA-633 */
        pInst->deviceType = CFONTZ_DEVICE_TYPE_CFA633;
        pInst->nHeight   = 2;
        pInst->nWidth    = 16;
        pInst->nLEDCount = 0;
        pInst->nFanCount = 0;
    }
    else {
        traceDebug(IILC_TRACE_CONTEXT, "Did not recognise device type");
    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}

int cfontz_UpdateFanPower(LP_CFONTZ_DEV_INSTANCE pInst)
{
    int rc = 0;
    int i = 0;

    LPPACKET pPacket = NULL, pReplyPacket = NULL;

    /* Sanity check */
    if (pInst == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* All fans have been set to sensible values already */
    rc = packetCreate(PACKET_TYPE_SET_FAN_POWER, &pPacket);
    if (rc) { goto end_of_function; }
    
    /* Write all fan data */
    for (i = 0; i < pInst->nFanCount; i++) {
        packetWriteData(pPacket, pInst->nFanPower[i]);
    }

    /* Send data to the display */
    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    /*
    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        printf("Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
    }
    */

end_of_function:

    /* Clean up resources */
    if (pPacket)      { (void) packetDispose(&pPacket); }
    if (pReplyPacket) { (void) packetDispose(&pReplyPacket); }

    return rc;


}

/**
  * \brief Takes the temperature reporting flags and updates the display.
  *
  * \param pInst Pointer to a valid device instance.
  *
  * \retval int 0 = success.
  */
int cfontz_UpdateTempReport(LP_CFONTZ_DEV_INSTANCE pInst)
{
    int rc = 0;
    int i = 0;

    LPPACKET pPacket = NULL, pReplyPacket = NULL;
    PACKET_TYPE replyType;

    IILC_TRACE_ENTRY(IILC_MODULE, T_cfontz_UpdateTempReport);

    /* Sanity check */
    if (pInst == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* All fans have been set to sensible values already */
    rc = packetCreate(PACKET_TYPE_SET_TEMP_REPORTING, &pPacket);
    if (rc) { goto end_of_function; }
    
    /* Write all fan data */
    for (i = 0; i < 4; i++) {
        unsigned int val = (pInst->tempReportMask & (0xFF << (i*8))) >> (i*8);
        packetWriteData(pPacket, (unsigned char) val);
        traceDebug(IILC_TRACE_CONTEXT, "Written data %u", val);
    }

    /* Send data to the display */
    rc = cfontz_WritePacket(pInst, pPacket, &pReplyPacket);
    if (rc) { goto end_of_function; }

    rc = packetGetType(pReplyPacket, &replyType);
    if (replyType != (PACKET_TYPE_SET_TEMP_REPORTING | 0x40)) {
        traceError(IILC_TRACE_CONTEXT, 
            "Bad packet ack received for temperature reporting (%d)\n", replyType);
        rc = -1;
        goto end_of_function;
    }

end_of_function:

    /* Clean up resources */
    if (pPacket)      { (void) packetDispose(&pPacket); }
    if (pReplyPacket) { (void) packetDispose(&pReplyPacket); }

    IILC_TRACE_EXIT_RC(rc);
    return rc;


}
