/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#if !defined(__CFONTZ_H__)
#define __CFONTZ_H__

#include "driver.h"
#include "packet.h"


typedef struct _tag_cfontz_device_instance * LP_CFONTZ_DEV_INSTANCE;


int cfontz_InitDevice(LP_CFONTZ_DEV_INSTANCE pInst);

/*
 * Some OS-specific stuff here for device initialisation, which includes:
 *
 * - How we refer to the OS device handle
 * - Macro which expands to the function used for initialisation
 * - A prototype for the above function
 */

#if defined(_WIN32)
#   define PHYSICAL_DEVICE_HANDLE       HANDLE
#   define NULL_PHYSICAL_DEVICE_HANDLE  0
#else
#   define PHYSICAL_DEVICE_HANDLE       int
#   define NULL_PHYSICAL_DEVICE_HANDLE  0
#endif



typedef enum _tag_cfontz_device_type
{
    CFONTZ_DEVICE_TYPE_UNKNOWN = 0,
    CFONTZ_DEVICE_TYPE_CFA635  = 1,
    CFONTZ_DEVICE_TYPE_CFA633  = 2
}
CFontzDeviceType;



/**
  * \brief Testing
  */

typedef struct _tag_cfontz_device_instance
{
    unsigned int nHeight;

    unsigned int nWidth;

    unsigned int nLEDCount;

    unsigned int nFanCount;

    PHYSICAL_DEVICE_HANDLE Device;

    /* ID of the RxThread */
    pthread_t        RxThreadID;

    /* Mutex to lock when waiting for an Rx packet */
    pthread_mutex_t  RxMutex;

    /* Semaphore on which to wait when waiting for an Rx packet */
    pthread_cond_t   RxSemaphore;

    /* Flag to indicate the Rx thread should stay running */
    unsigned int     RxRunning;

    /* Mutex used to sync up running status */
    pthread_mutex_t  RunningMutex;

    /* The packet received in reply */
    LPPACKET         pReplyPacket;

    /** Structure which contains the function pointers for the various callbacks */
    struct {
        pfnDrvCBFanSpeed FanSpeed;
        pfnDrvCBTemp     Temp;
        pfnDrvCBKeyEvent KeyEvent;
    } pfn;

    /** Structure which contains the context data to supply for the various callbacks */
    struct {
        void * FanSpeed;
        void * Temp;
        void * KeyEvent;
    } context;

    /** Array containing the current power levels of all the fans.
      * This is necessary because the fans are all updated together in a 
      * single command and the interface dictates we update fan speeds
      * on a per-fan basis.
      */
    unsigned int * nFanPower;

    CFontzDeviceType deviceType;

    /* Current fan reporting mask */
    unsigned int fanReportMask;

    /* Current temperature sensor reporting mask */
    unsigned int tempReportMask;

    /* Current mask used to report key presses */
    unsigned int keyPressMask;

    /* Current mask used to report key releases */
    unsigned int keyReleaseMask;

    /** ???? */
    char *     pszPort;

    /** Flag to indicate the display has been started */
    unsigned int bStarted;

    /** Flag to indicate a SCAB board is present */
    unsigned int bSCABPresent;

} CFONTZ_DEV_INSTANCE;


/*
 * Trace constants
 */

#define T_driver_Create                 0x2A
#define T_cfontz_InitDevice             0x2B
#define T_cfontz_InitDevice_unix        0x2C
#define T_cfontz_InitDevice_w32         0x2D
#define T_parserDeviceUseLED            0x2E
#define T_cfontz_InitDeviceType         0x2F
#define T_cfontz_UpdateTempReport       0x30

#endif /* __CFONTZ_H__ */


