/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <memory.h>

#include "lcd.h"
#include "packet.h"
#include "trace.h"


#define IILC_MODULE MODULE_TYPE_PACKET


/* Definition of a packet */
typedef struct _tag_packet {

    /* Command packet type */
    unsigned char type;

    /* Current data length */
    unsigned char data_length;

    /* The data to send as a supplement */
    unsigned char data[MAX_DATA_LENGTH];

    /* CRC check value - initialised at send time */
    unsigned short CRC;

    /* Cursor for when we're reading data */
    unsigned char cursor;

} PACKET;

static int calculateCRC(LPPACKET pPacket, unsigned short * pCRC)
{
    /* CRC lookup table to avoid bit-shifting loops */
    static const unsigned short crcLookupTable[256] = {
        0x0000, 0x1189, 0x2312, 0x329B, 0x4624, 0x57AD, 0x6536, 0x74BF,
        0x8C48, 0x9DC1, 0xAF5A, 0xBED3, 0xCA6C, 0xDBE5, 0xE97E, 0xF8F7,
        0x1081, 0x0108, 0x3393, 0x221A, 0x56A5, 0x472C, 0x75B7, 0x643E,
        0x9CC9, 0x8D40, 0xBFDB, 0xAE52, 0xDAED, 0xCB64, 0xF9FF, 0xE876,
        0x2102, 0x308B, 0x0210, 0x1399, 0x6726, 0x76AF, 0x4434, 0x55BD,
        0xAD4A, 0xBCC3, 0x8E58, 0x9FD1, 0xEB6E, 0xFAE7, 0xC87C, 0xD9F5,
        0x3183, 0x200A, 0x1291, 0x0318, 0x77A7, 0x662E, 0x54B5, 0x453C,
        0xBDCB, 0xAC42, 0x9ED9, 0x8F50, 0xFBEF, 0xEA66, 0xD8FD, 0xC974,
        0x4204, 0x538D, 0x6116, 0x709F, 0x0420, 0x15A9, 0x2732, 0x36BB,
        0xCE4C, 0xDFC5, 0xED5E, 0xFCD7, 0x8868, 0x99E1, 0xAB7A, 0xBAF3,
        0x5285, 0x430C, 0x7197, 0x601E, 0x14A1, 0x0528, 0x37B3, 0x263A,
        0xDECD, 0xCF44, 0xFDDF, 0xEC56, 0x98E9, 0x8960, 0xBBFB, 0xAA72,
        0x6306, 0x728F, 0x4014, 0x519D, 0x2522, 0x34AB, 0x0630, 0x17B9,
        0xEF4E, 0xFEC7, 0xCC5C, 0xDDD5, 0xA96A, 0xB8E3, 0x8A78, 0x9BF1,
        0x7387, 0x620E, 0x5095, 0x411C, 0x35A3, 0x242A, 0x16B1, 0x0738,
        0xFFCF, 0xEE46, 0xDCDD, 0xCD54, 0xB9EB, 0xA862, 0x9AF9, 0x8B70,
        0x8408, 0x9581, 0xA71A, 0xB693, 0xC22C, 0xD3A5, 0xE13E, 0xF0B7,
        0x0840, 0x19C9, 0x2B52, 0x3ADB, 0x4E64, 0x5FED, 0x6D76, 0x7CFF,
        0x9489, 0x8500, 0xB79B, 0xA612, 0xD2AD, 0xC324, 0xF1BF, 0xE036,
        0x18C1, 0x0948, 0x3BD3, 0x2A5A, 0x5EE5, 0x4F6C, 0x7DF7, 0x6C7E,
        0xA50A, 0xB483, 0x8618, 0x9791, 0xE32E, 0xF2A7, 0xC03C, 0xD1B5,
        0x2942, 0x38CB, 0x0A50, 0x1BD9, 0x6F66, 0x7EEF, 0x4C74, 0x5DFD,
        0xB58B, 0xA402, 0x9699, 0x8710, 0xF3AF, 0xE226, 0xD0BD, 0xC134,
        0x39C3, 0x284A, 0x1AD1, 0x0B58, 0x7FE7, 0x6E6E, 0x5CF5, 0x4D7C,
        0xC60C, 0xD785, 0xE51E, 0xF497, 0x8028, 0x91A1, 0xA33A, 0xB2B3,
        0x4A44, 0x5BCD, 0x6956, 0x78DF, 0x0C60, 0x1DE9, 0x2F72, 0x3EFB,
        0xD68D, 0xC704, 0xF59F, 0xE416, 0x90A9, 0x8120, 0xB3BB, 0xA232,
        0x5AC5, 0x4B4C, 0x79D7, 0x685E, 0x1CE1, 0x0D68, 0x3FF3, 0x2E7A,
        0xE70E, 0xF687, 0xC41C, 0xD595, 0xA12A, 0xB0A3, 0x8238, 0x93B1,
        0x6B46, 0x7ACF, 0x4854, 0x59DD, 0x2D62, 0x3CEB, 0x0E70, 0x1FF9,
        0xF78F, 0xE606, 0xD49D, 0xC514, 0xB1AB, 0xA022, 0x92B9, 0x8330,
        0x7BC7, 0x6A4E, 0x58D5, 0x495C, 0x3DE3, 0x2C6A, 0x1EF1, 0x0F78
    };

    /* Initialise the CRC field */
    register unsigned short newCRC = 0xFFFF;

    /* Temporary copy of the length field */
    unsigned short len;

    /* Pointer to the buffer */
    unsigned char * bufPtr;

    /* Parameter sanity check */
    IILC_ASSERT(pPacket != NULL);
    IILC_ASSERT(pCRC != NULL);

    /* Convenient local vars */
    len    = pPacket->data_length;
    bufPtr = pPacket->data;

    /* Do CRC operation on packet type field (newCRC >> 8 = 0xFF) */
    newCRC = 0xFF ^ crcLookupTable[(newCRC ^ pPacket->type) & 0xFF];

    /* Include the data length in the operation */
    newCRC = (newCRC >> 8) ^ crcLookupTable[(newCRC ^ pPacket->data_length) & 0xFF];

    /* Perform the bit manipulation on the data region */
    while (len--) {
        newCRC = (newCRC >> 8) ^ crcLookupTable[(newCRC ^ *bufPtr++) & 0xff];
    }

    /* Do one's complement on this & return */
    *pCRC = ~newCRC;

    /* Return OK */
    return 0;
}


int packetCreate(PACKET_TYPE type, LPPACKET * ppPacket)
{
    int      rc      = 0;
    int      i       = 0;
    LPPACKET pPacket = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_packetCreate);
    traceDebug(IILC_TRACE_CONTEXT,
        "type " T_FMT_INT ", ppPacket " T_FMT_PTR, type, ppPacket);

    IILC_ASSERT(ppPacket != NULL);



    /* Initialise the memory */
    pPacket = (LPPACKET) malloc(sizeof(PACKET));
    if (pPacket == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* Initialise the data inside */
    pPacket->type = type;
    pPacket->data_length = 0;
    pPacket->CRC = 0xFFFF;
    pPacket->cursor = 0;

    /* Clear the data field */
    for (i = 0; i < MAX_DATA_LENGTH; i++) {
        pPacket->data[i] = 0;
    }

    /* Return to the user */
    *ppPacket = pPacket;

end_of_function:

    traceDebug(IILC_TRACE_CONTEXT, "rc " T_FMT_INT, rc);
    IILC_TRACE_EXIT;
    return rc;
}

int packetDispose(LPPACKET * ppPacket)
{
    IILC_ASSERT(ppPacket != NULL);

    if (ppPacket == NULL) {
        abort();
    }

    if (*ppPacket == NULL) {
        abort();
    }

    memset(*ppPacket, 0x55, sizeof(PACKET));

    /* Free the memory allocated */
    free(*ppPacket);

    /* Release user's reference */
    *ppPacket = NULL;

    /* All OK */
    return 0;
}

int packetWriteData(LPPACKET pPacket, unsigned char data)
{
    int rc = 0;

    IILC_ASSERT(pPacket != NULL);

    /* Check packet isn't full */
    if (pPacket->data_length >= MAX_DATA_LENGTH) {
        rc = ERR_PACKET_FULL;
        goto end_of_function;
    }

    /* Write data to packet */
    pPacket->data[pPacket->data_length++] = data;

end_of_function:

    return rc;
}

/* Writes a string as bytes to the packet. Note this string is NOT
 * null-terminated as a string may validly contain a NUL character.
 */
int packetWriteString(LPPACKET pPacket, char * strData, int strDataLen)
{
    int rc = 0;

    IILC_ASSERT(pPacket != NULL);
    IILC_ASSERT(strData != NULL);
    IILC_ASSERT(strDataLen >= 0);

    /* Check packet isn't full */
    if (pPacket->data_length + strDataLen > MAX_DATA_LENGTH) {
        rc = ERR_PACKET_FULL;
        goto end_of_function;
    }

    /* Write data to the packet */
    memcpy(pPacket->data + pPacket->data_length, strData, strDataLen);

    /* Increase packet size accordingly */
    pPacket->data_length += strDataLen;

end_of_function:

    return rc;
}

int packetReadData(LPPACKET pPacket, unsigned char * data)
{
    int rc = 0;

    IILC_ASSERT(pPacket != NULL);
    IILC_ASSERT(data != NULL);

    /* Check for end of packet */
    if (pPacket->cursor >= pPacket->data_length) {
        rc = ERR_END_OF_PACKET;
        *data = 0;
        goto end_of_function;
    }

    /* Read data from packet & advance cursor */
    *data = pPacket->data[pPacket->cursor++];

end_of_function:

    /* Return any errors */
    return rc;
}


int packetReadReset(LPPACKET pPacket)
{
    IILC_ASSERT(pPacket != NULL);

    pPacket->cursor = 0;

    /* All OK */
    return 0;
}


int packetFlatten(LPPACKET pPacket, unsigned char * buf, int * pDataLen)
{
    int            rc  = 0;
    unsigned short CRC = 0xFFFF;
    
    IILC_ASSERT(pPacket != NULL);
    IILC_ASSERT(buf != NULL);
    IILC_ASSERT(pDataLen != NULL);

    /* Check there's enough room */
    if ( *pDataLen < pPacket->data_length + 4 ) {
        rc = ERR_BUFFER_TOO_SMALL;
        goto end_of_function;
    }

	/* Calculate our CRC */
	rc = calculateCRC(pPacket, &CRC);
    if (rc) { goto end_of_function; }
	
	/* Copy data into buffer */
	buf[0] = pPacket->type;
	buf[1] = pPacket->data_length;
    memcpy(buf + 2, &(pPacket->data), buf[1]);

	/* Put in CRC LSB first */
	buf[buf[1]+2] = CRC & 0xFF;
	buf[buf[1]+3] = CRC >> 8;

	/* Total size of buffer is data + 4 */
	*pDataLen = pPacket->data_length + 4;

end_of_function:

    return rc;
}

int packetInflate(unsigned char * buf, int len, int * pBytesRead, LPPACKET * ppPacket)
{
    int rc        = 0;
    int bytesRead = 0;
    unsigned short CRC       = 0;
   
    LPPACKET pPacket = NULL;

    /* Construct memory region - keeps things cleaner in the loop */
    pPacket = (LPPACKET) malloc(sizeof(PACKET));
    if (pPacket == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    while (bytesRead < len) {

        /* Check there's enough data for a packet - at least 4 bytes */
        if ((len - bytesRead) < 4) {
            rc = ERR_INSUFFICIENT_DATA;
            break;
        }

        /* Does this look like a packet from the display? */
        if (buf[bytesRead] < 0x40) {            
            bytesRead++;    /* No - skip ahead in buffer */
            continue;
        }

        /* Is this a sensible data length? */
        if (buf[bytesRead+1] > MAX_DATA_LENGTH) {
            bytesRead++;    /* No - skip ahead in buffer */
            continue;
        }

        /* Is there enough data in the buffer for this packet? */
        if (bytesRead + buf[bytesRead+1] + 4 > len) {
            /* Not enough data read at present */
            rc = ERR_INSUFFICIENT_DATA;
            break;
        }

        /* Initialise the data inside */
        pPacket->type = buf[bytesRead];
        pPacket->data_length = buf[bytesRead+1];
        pPacket->cursor = 0;

        /* Copy in the data field */
        memcpy(pPacket->data, buf + bytesRead + 2, pPacket->data_length);

        /* Get the CRC (LSB first) */
        pPacket->CRC =  buf[bytesRead + 2 + pPacket->data_length];
        pPacket->CRC += buf[bytesRead + 3 + pPacket->data_length] << 8;

        /* Check the CRC */
        rc = calculateCRC(pPacket, &CRC);
        if (rc) {
            printf("Error calculating CRC\n");
            break;
        }

        /* If CRC is correct, we have read a complete packet!! */
        if (CRC != pPacket->CRC) {
            printf("Rejected on bad CRC\n");
            bytesRead++;
            continue;
        }
        else {
            bytesRead += 4 + pPacket->data_length;
            break;
        }
    }

end_of_function:

    /* Let calling function know how many bytes we read */
    if (pBytesRead != NULL) {
        *pBytesRead = bytesRead;
    }

    if (!rc) {
        *ppPacket = pPacket;
    }
    else {
        if (pPacket != NULL) {
            free(pPacket);
        }
    }

    return rc;
}

int packetGetClass(LPPACKET pPacket)
{
    return (pPacket->type & 0xC0) >> 6;
}

int packetGetType(LPPACKET pPacket, PACKET_TYPE * pType)
{
    *pType = pPacket->type;
    return 0;
}

