/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <time.h>
#include <string.h>

#if !defined(_WIN32)
#   include <strings.h>
#endif

#include "lcd.h"
#include "module.h"
#include "screen.h"
#include "trace.h"

#define IILC_MODULE MODULE_TYPE_DISPLAY

LCD_API int ModuleCreate(int handle,
                         void ** context,
                         LPMODULE_FN_ARRAY pFnArray);

LCD_API int ModuleDestroy(void * context);


LCD_API int GetDisplayData(void * context,
                           int row,
                           int nWidth,
                           char * buf);


typedef struct _tag_screen_fans
{
    /** Number of fans being monitored */
    unsigned int nFanCount;

    /** Array of unsigned ints to contain rpm values */
    unsigned int * rpmList;

    /** Scratch space to keep our data stored */
    char ** pszOutput;

} SCREEN_FANS, * LPSCREEN_FANS;


/* ----- Internal function definitions ----- */


void fanSpeedReport(void * context, 
                    unsigned int fan_id, 
                    unsigned int fan_speed,
                    const char * pszName);


/* ----- Implementation ----- */



int ModuleCreate(int handle, void ** context, LPMODULE_FN_ARRAY pFnArray)
{
    int rc = 0;

    LPSCREEN_FANS pInst = NULL;

    /* Create our own scratch space */
    pInst = malloc(sizeof(SCREEN_FANS));
    if (pInst == NULL) {
        printf("malloc() failed in Fans module create");
        rc = -1;
        goto end_of_function;
    }

    /* Default to looking after four fans */
    pInst->nFanCount = 4;

    /* Zero out the array contents */
    pInst->rpmList = calloc(pInst->nFanCount, sizeof(unsigned int));

    /* --- Populate the function array data --- */

    /* We can use fn array version 1 */
    pFnArray->version = 1;

    /* Setup our various functions */
    pFnArray->pfnTextData = GetDisplayData;
    pFnArray->pfnFanSpeedReport = fanSpeedReport;

    /* --- */

    /* Return context data back to the caller */
    *context = pInst;

end_of_function:

    return rc;
}


int ModuleDestroy(void * context)
{
    LPSCREEN_FANS pInst = (LPSCREEN_FANS) context;

    free(pInst);

    return 0;
}


void temperatureReport(void * context, unsigned int sensor_id, double degrees_c, const char * pszName)
{
    IILC_TRACE_ENTRY(IILC_MODULE, T_temperatureReport);

    traceInfo(IILC_TRACE_CONTEXT, "(FANS) Temperature report for sensor [%10s] at %.0fC", pszName, degrees_c);

    IILC_TRACE_EXIT;
}



void fanSpeedReport(void * context, unsigned int fan_id, unsigned int fan_speed, const char * pszName)
{
    LPSCREEN_FANS pInst = NULL;
        
    IILC_TRACE_ENTRY(IILC_MODULE, T_fanSpeedReport);

    /* Grab the instance data */
    pInst = (LPSCREEN_FANS) context;

    switch (fan_speed) {
    case 0:
    case 1:
    case 0xFFFFFFFF:
        pInst->rpmList[fan_id] = fan_speed;
        break;

    /* Save the RPM data in some shared storage (round to nearest 100) */
    default:
        pInst->rpmList[fan_id] = fan_speed - ((fan_speed - 50) % 100) + 50;
        break;
    }

    traceInfo(IILC_TRACE_CONTEXT, "(FANS) Fan speed report of sensor    [%10s] at %u RPM", pszName, fan_speed);
    
    IILC_TRACE_EXIT;
}



int GetDisplayData(void * context, int row, int nWidth, char * buf)
{
    /* Return code */
    int rc = 0;

    /* This module instance */
    LPSCREEN_FANS pInstance = (LPSCREEN_FANS) context;

    switch (pInstance->rpmList[row]) {

    case 0: /* Stopped */
        sprintf(buf, "FAN %u: STOPPED", row);
        break;

    case 1: /* Below measurable limits */
        sprintf(buf, "FAN %u: SLOW", row);
        break;

    case 0xFFFFFFFF: /* Over-limit */
        sprintf(buf, "FAN %u: ---- RPM", row);
        break;

    default:
        sprintf(buf, "FAN %u: %4u RPM", row, pInstance->rpmList[row]);
        break;
    }

    return rc;
}
