/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <string.h>
#include "config.h"
#include "module.h"
#include "trace.h"
#include "vector.h"

#define IILC_MODULE MODULE_TYPE_APP



/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/


/**
  * A top-level configuration object. Each instance of the server has its own
  * configuration file.
  */
typedef struct _tag_config_file
{
    /* ----- Data held at parse time ----- */

    /** Name of the file currently being parsed */
    char * pszFileName;

    /** The list of displays (type LPCONFIG_DISPLAY) this file is configuring */
    LPVECTOR pDisplays;

    /* ----- Data calculated during verify stage ----- */



} CONFIG_FILE;


/**
  * A configuration file may contain one or more displays defined within.
  */
typedef struct _tag_config_display
{
    /* ----- Data held at parse time ----- */

    const char * pszName;


    LPCONFIG_DEVICE pDevice;


    LPVECTOR pModuleList;

    LPVECTOR pScreenList;


    /* ----- Data calculated during verify stage ----- */




    /* ----- Child data ----- */

    /** A display is run by mutiple modules */
//    LPVECTOR pModules;

    /** Each display can show multiple screens */
//    LPVECTOR pScreens;

    /* -- Current screen being configured -- */
//    LPSCREEN pCurrentScreen;


    /* ----- Config data ----- */

    /** The single device we are controlling */
//    LPCONFIG_DEVICE pConfigDevice;

    /** Name of this display */
//    char * pszName;
    
    /* ----- Runtime data ----- */

    /** Pointer to the display to use */
//    LPDISPLAY pDisplay;

} CONFIG_DISPLAY;


/**
  * Each display is connected one and one only physical device.
  */
typedef struct _tag_config_device
{
    /* ----- Data held at parse time ----- */

    /** A vector of LPCONFIG_OBJECT elements listing resources available */
    LPVECTOR pObjectList;

    /** Name of this device */
    char * pszName;

    char * pszPort;

    char * pszClass;

    /* ----- Data calculated during verify stage ----- */


} CONFIG_DEVICE;


/**
  * The behaviour of each display is controlled by one or more plugin modules.
  */
typedef struct _tag_config_module
{
    /* ----- Data held at parse time ----- */

    /** Name of the module */
    const char * pszModuleName;

    /** Name of the library to load */
    const char * pszLibraryName;

    /** A vector of LPCONFIG_OBJECT elements listing objects we wish to receive reports for */
    LPVECTOR pReportList;

    /** A vector of LPCONFIG_OBJECT elements listing objects we wish to control */
    LPVECTOR pControlList;

    /* ----- Data calculated during verify stage ----- */



} CONFIG_MODULE;



/**
  * What is output on the display is defined by the make-up of one or more screens.
  */
typedef struct _tag_config_screen
{
    char * pszScreen;

    LPVECTOR pWindowList;


}
CONFIG_SCREEN;


typedef struct _tag_config_window
{
    /* ----- Data held at parse time ----- */

    int x;

    int y;

    int width;

    int height;

    char * pszModule;

    /* ----- Data calculated during verify stage ----- */

    LPCONFIG_MODULE pModule;
}
CONFIG_WINDOW;


/**
  * Various parameters about the display can be monitored.
  */
typedef struct _tag_config_object
{
    enum displayObjectType type;

    /** Internal ID as reported to the physical device */
    unsigned int physical_id;

    /** Name as used in the config file */
    const char * pszName;

    /** The module which is allowed to control this object */
    LPCONFIG_MODULE pControllingModule;

    /** The ID which the module will use to control this object */
    unsigned int nModuleControlAsID;

    /** A vector containing modules who wish to receive notifications on this object */
    LPVECTOR pReportModulesList;

    /** A vector of 'int' that shows the ID against which modules should receive reports */
    LPVECTOR pReportIDList;

} CONFIG_OBJECT;



/****************************************************************************/
/*                                                                          */
/* Config object function prototypes                                       */
/*                                                                          */
/****************************************************************************/


static int configObjectCreate(LPCONFIG_OBJECT * ppObject, 
                              enum displayObjectType type, 
                              int physical_id, 
                              const char * pszName);



/****************************************************************************/
/*                                                                          */
/* Config file functions                                                    */
/*                                                                          */
/****************************************************************************/


int configFileCreate(LPCONFIG_FILE * ppFile, const char * pszName)
{
    int rc = 0;

    LPCONFIG_FILE pFile = NULL;

    /* Alloc storage */
    pFile = (LPCONFIG_FILE) calloc(1, sizeof(CONFIG_FILE));

    /* Initialise */
    pFile->pszFileName = strdup(pszName);
    vectorCreate(&(pFile->pDisplays));

    /* Return to caller */
    *ppFile = pFile;

    return rc;
}


int configFileGetDisplayCount(LPCONFIG_FILE pConfig, int * pnCount)
{
    int rc = 0;
    int nCount = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetDisplayCount);
    traceDebug(IILC_TRACE_CONTEXT, "pConfig %p, pnCount %p", pConfig, pnCount);

    if (pnCount == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    rc = vectorGetSize(pConfig->pDisplays, &nCount);
    if (rc) {
        goto end_of_function;
    }

    *pnCount = nCount;
    traceDebug(IILC_TRACE_CONTEXT, "*pnCount %d", *pnCount);

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}



int configFileGetDisplay(LPCONFIG_FILE pFile, int nIndex, LPCONFIG_DISPLAY * ppDisplay)
{
    int rc = 0;

    vectorGetElementAt(pFile->pDisplays, nIndex, (void **) ppDisplay);

    return rc;
}



int configFileAddDisplay(LPCONFIG_FILE pFile, LPCONFIG_DISPLAY pDisplay)
{
    int rc = 0;

    vectorAdd(pFile->pDisplays, pDisplay);

    return rc;
}

int configFileValidate(LPCONFIG_FILE pFile)
{
    int rc = 0;

    return rc = 0;
}



int configFileDispose(LPCONFIG_FILE * ppInst)
{
    LPCONFIG_FILE pConfig = *ppInst;
    int count = 0;
    int i = 0;
    int rc = 0;

    vectorGetSize(pConfig->pDisplays, &count);

    /* Destory each display */
    for (i = 0; (i < count) && (rc == 0); i++) {

        LPCONFIG_DISPLAY pDisplay = NULL;

        rc = vectorGetElementAt(pConfig->pDisplays, i, (void **) &pDisplay);

        /* Remove from vector */
        vectorRemove(pConfig->pDisplays, pDisplay);

        /* Dispose of the display */
        configDisplayDispose(&pDisplay);
    }

    /* Dispose of vector list */
    vectorDestroy(&(pConfig->pDisplays));

    if (pConfig->pszFileName) free(pConfig->pszFileName);

    /* Release structure's memory */
    free(pConfig);

    /* Wipe user's reference */
    *ppInst = NULL;

    return 0;
}


/****************************************************************************/
/*                                                                          */
/* Config display functions                                                 */
/*                                                                          */
/****************************************************************************/


int configDisplayCreate(LPCONFIG_DISPLAY * ppDisplay, const char * pszDisplay)
{
    int rc = 0;
    LPCONFIG_DISPLAY pDisplay = NULL;

    /* Allocate storage */
    pDisplay = (LPCONFIG_DISPLAY) calloc(1, sizeof(CONFIG_DISPLAY));
    pDisplay->pszName = strdup(pszDisplay);
    vectorCreate(&(pDisplay->pModuleList));
    vectorCreate(&(pDisplay->pScreenList));

    /* Return to caller */
    *ppDisplay = pDisplay;

    return rc;
}


int configDisplayGetDevice(LPCONFIG_DISPLAY pDisplay, LPCONFIG_DEVICE * ppDevice)
{
    int rc = 0;


    *ppDevice = pDisplay->pDevice;


    return rc;
}


int configDisplaySetDevice(LPCONFIG_DISPLAY pDisplay, LPCONFIG_DEVICE pDevice)
{
    int rc = 0;

    pDisplay->pDevice = pDevice;

    return rc;
}

int configDisplayAddModule(LPCONFIG_DISPLAY pDisplay, LPCONFIG_MODULE pModule)
{
    int rc = 0;

    vectorAdd(pDisplay->pModuleList, pModule);

    return rc;
}

int configDisplayGetModuleCount(LPCONFIG_DISPLAY pDisplay, int * pnCount)
{
    int rc = 0;

    vectorGetSize(pDisplay->pModuleList, pnCount);

    return rc;
}

int configDisplayGetModule(LPCONFIG_DISPLAY pDisplay, int index, LPCONFIG_MODULE * ppModule)
{
    int rc = 0;

    vectorGetElementAt(pDisplay->pModuleList, index, (void **) ppModule);

    return rc;
}

int configDisplayAddScreen(LPCONFIG_DISPLAY pDisplay, LPCONFIG_SCREEN pScreen)
{
    int rc = 0;

    vectorAdd(pDisplay->pScreenList, pScreen);

    return rc;
}

int configDisplayGetScreenCount(LPCONFIG_DISPLAY pDisplay, int * pnCount)
{
    int rc = 0;

    vectorGetSize(pDisplay->pScreenList, pnCount);

    return rc;
}

int configDisplayGetScreen(LPCONFIG_DISPLAY pDisplay, int nIndex, LPCONFIG_SCREEN * ppScreen)
{
    int rc = 0;

    vectorGetElementAt(pDisplay->pScreenList, nIndex, (void **) ppScreen);

    return rc;
}

int configDisplayDispose(LPCONFIG_DISPLAY * ppDisplay)
{
    int rc = 0;
    LPCONFIG_DISPLAY pDisplay = NULL;

    pDisplay = *ppDisplay;

    configDeviceDispose(&(pDisplay->pDevice));

    vectorDestroy(&(pDisplay->pModuleList));
    vectorDestroy(&(pDisplay->pScreenList));

    free((char *) pDisplay->pszName);

    free(pDisplay);

    return rc;
}


/****************************************************************************/
/*                                                                          */
/* Config device functions                                                  */
/*                                                                          */
/****************************************************************************/




int configDeviceCreate(LPCONFIG_DEVICE * ppDevice, const char * pszDevice)
{
    int rc = 0;
    LPCONFIG_DEVICE pDevice = NULL;

    /* Create new device & init */
    pDevice = (LPCONFIG_DEVICE) calloc(1, sizeof(CONFIG_DEVICE));
    if (pDevice == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* Initialise the structure */
    pDevice->pszName = strdup(pszDevice);
    if (pDevice->pszName == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    vectorCreate(&(pDevice->pObjectList));

    /* Return to the user */
    *ppDevice = pDevice;

end_of_function:

    return rc;
}




int configDeviceUseObject(LPCONFIG_DEVICE pDevice, enum displayObjectType type, 
                          int physical_id, const char * pszName)
{
    int rc = 0;
    LPCONFIG_OBJECT pObject = NULL;

    /* Create this new object */
    configObjectCreate(&pObject, type, physical_id, pszName);

    /* Check this sensor ID has not been used */


    /* Add this object to the device */
    vectorAdd(pDevice->pObjectList, (void *) pObject);

    return rc;
}

int configDeviceSetClass(LPCONFIG_DEVICE pDevice, const char * pszClass)
{
    int rc = 0;

    pDevice->pszClass = strdup(pszClass);

    return rc;
}

int configDeviceSetPort(LPCONFIG_DEVICE pDevice, const char * pszPort)
{
    int rc = 0;

    pDevice->pszPort = strdup(pszPort);

    return rc;
}

int configDeviceGetClass(LPCONFIG_DEVICE pDevice, char ** ppszClass)
{
    int rc = 0;

    *ppszClass = pDevice->pszClass;

    return rc;
}


int configDeviceGetPort(LPCONFIG_DEVICE pDevice, char ** ppszPort)
{
    int rc = 0;

    *ppszPort = pDevice->pszPort;

    return rc;
}

int configDeviceDispose(LPCONFIG_DEVICE * ppDevice)
{
    int rc = 0;
    LPCONFIG_DEVICE pDevice = NULL;

    pDevice = *ppDevice;

    vectorDestroy(&(pDevice->pObjectList));

    if (pDevice->pszClass) free(pDevice->pszClass);
    if (pDevice->pszName)  free(pDevice->pszName);
    if (pDevice->pszPort)  free(pDevice->pszPort);

    free(pDevice);

    return rc;
}

/****************************************************************************/
/*                                                                          */
/* Config module functions                                                  */
/*                                                                          */
/****************************************************************************/

int configModuleCreate(LPCONFIG_MODULE * ppModule, const char * pszModule)
{
    int rc = 0;
    LPCONFIG_MODULE pModule = NULL;

    /* Create the new struct */
    pModule = (LPCONFIG_MODULE) calloc(1, sizeof(CONFIG_MODULE));
    if (pModule == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* Init */
    pModule->pszModuleName = strdup(pszModule);
    if (pModule->pszModuleName == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    vectorCreate(&(pModule->pControlList));
    vectorCreate(&(pModule->pReportList));

    /* Return to the user */
    *ppModule = pModule;

end_of_function:

    return rc;
}


int configModuleGetLibrary(LPCONFIG_MODULE pModule, const char ** ppszName)
{
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_configGetModuleName);
    traceDebug(IILC_TRACE_CONTEXT, "pModule %p, ppszName %p", pModule, ppszName);

    if (pModule == NULL) {
        traceError(IILC_TRACE_CONTEXT, "pModule == NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    if (ppszName == NULL) {
        traceError(IILC_TRACE_CONTEXT, "ppszName == NULL");
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    *ppszName = pModule->pszLibraryName;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int configModuleReportObject(LPCONFIG_MODULE pModule, enum displayObjectType type, 
                             int object_id, const char * pszReportAs)
{
    int rc = 0;
    LPCONFIG_OBJECT pObject = NULL;

    /* Create a new object */
    configObjectCreate(&pObject, type, object_id, pszReportAs);

    /* Add this object to the list of reports the module requires */
    vectorAdd(pModule->pReportList, pObject);

    return rc;
}


int configModuleControlObject(LPCONFIG_MODULE pModule, enum displayObjectType type, int object_id)
{
    int rc = 0;
    LPCONFIG_OBJECT pObject = NULL;

    /* Create a new object */
    configObjectCreate(&pObject, type, object_id, NULL);

    /* Add this object to the list of objects this module controls */
    vectorAdd(pModule->pControlList, pObject);

    return rc;
}

int configModuleVirtualObject(LPCONFIG_MODULE pModule, int module_id, int device_id, const char * pszObject)
{
    int rc = 0;

    return rc;
}


int configModuleSetLibrary(LPCONFIG_MODULE pModule, const char * pszLibrary)
{
    int rc = 0;

    pModule->pszLibraryName = strdup(pszLibrary);

    return rc;
}


int configModuleGetReportCount(LPCONFIG_MODULE pModule, int * pnCount)
{
    int rc = 0;

    vectorGetSize(pModule->pReportList, pnCount);

    return rc;
}

int configModuleGetReportObject(LPCONFIG_MODULE pModule, int index, LPCONFIG_OBJECT * ppObject)
{
    int rc = 0;

    vectorGetElementAt(pModule->pReportList, index, (void **) ppObject);

    return rc;
}










/****************************************************************************/
/*                                                                          */
/* Config screen functions                                                  */
/*                                                                          */
/****************************************************************************/


int configScreenCreate(LPCONFIG_SCREEN * ppScreen, const char * pszScreen)
{
    int rc = 0;
    LPCONFIG_SCREEN pScreen = NULL;

    pScreen = (LPCONFIG_SCREEN) calloc(1, sizeof(CONFIG_SCREEN));
    vectorCreate(&(pScreen->pWindowList));
    pScreen->pszScreen = strdup(pszScreen);

    *ppScreen = pScreen;

    return rc;
}


int configScreenShowModule(LPCONFIG_SCREEN pScreen, const char * pszModule, 
                           int size_x, int size_y, int pos_x, int pos_y)
{
    int rc = 0;
    LPCONFIG_WINDOW pWindow = NULL;

    /* Create a new window using this data */
    configWindowCreate(&pWindow, size_x, size_y, pos_x, pos_y, pszModule);

    /* Add this window to the screen */
    vectorAdd(pScreen->pWindowList, pWindow);

    return rc;
}

int configScreenGetWindowCount(LPCONFIG_SCREEN pScreen, int * pnCount)
{
    int rc = 0;

    vectorGetSize(pScreen->pWindowList, pnCount);

    return rc;
}

int configScreenGetWindow(LPCONFIG_SCREEN pScreen, int nIndex, LPCONFIG_WINDOW * ppWindow)
{
    int rc = 0;

    vectorGetElementAt(pScreen->pWindowList, nIndex, (void **) ppWindow);

    return rc;
}

int configScreenDispose(LPCONFIG_SCREEN * ppScreen)
{
    int rc = 0;
    LPCONFIG_SCREEN pScreen = NULL;

    pScreen = *ppScreen;

    if (pScreen->pszScreen != NULL) free(pScreen->pszScreen);
    if (pScreen->pWindowList != NULL) {

        int i, nCount = 0;

        /* Cleanly dispose of all child windows */
        vectorGetSize(pScreen->pWindowList, &nCount);
        for (i = 0; i < nCount; i++) {
            LPCONFIG_WINDOW pWindow = NULL;
            vectorGetElementAt(pScreen->pWindowList, i, (void **) &pWindow);
            configWindowDispose(&pWindow);
        }

        /* Clean up the internal list */
        vectorDestroy(&(pScreen->pWindowList));
    }

    free(pScreen);

    return rc;
}



/****************************************************************************/
/*                                                                          */
/* Config window functions                                                  */
/*                                                                          */
/****************************************************************************/

int configWindowCreate(LPCONFIG_WINDOW * ppWindow, int width, int height, 
                       int x, int y, const char * pszModule)
{
    int rc = 0;
    LPCONFIG_WINDOW pWindow = NULL;

    pWindow = (LPCONFIG_WINDOW) calloc(1, sizeof(CONFIG_WINDOW));
    pWindow->width     = width;
    pWindow->height    = height;
    pWindow->x         = x;
    pWindow->y         = y;
    pWindow->pszModule = strdup(pszModule);

    *ppWindow = pWindow;

    return rc;
}

int configWindowGetLocation(LPCONFIG_WINDOW pWindow, int * pnX, int * pnY)
{
    int rc = 0;

    *pnX = pWindow->x;
    *pnY = pWindow->y;

    return rc;
}

int configWindowGetSize(LPCONFIG_WINDOW pWindow, int * pnWidth, int * pnHeight)
{
    int rc = 0;

    *pnWidth  = pWindow->width;
    *pnHeight = pWindow->height;

    return rc;
}

int configWindowGetModuleName(LPCONFIG_WINDOW pWindow, const char ** ppszModuleName)
{
    int rc = 0;

    *ppszModuleName = pWindow->pszModule;

    return rc;
}

int configWindowDispose(LPCONFIG_WINDOW * ppWindow)
{
    int rc = 0;

    free(*ppWindow);

    return rc;
}

/****************************************************************************/
/*                                                                          */
/* Config object functions                                                  */
/*                                                                          */
/****************************************************************************/

int configObjectCreate(LPCONFIG_OBJECT * ppObject, enum displayObjectType type, 
                       int physical_id, const char * pszName)
{
    int rc = 0;
    LPCONFIG_OBJECT pObject = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserDeviceCreateObject);

    /* Alloc and init the structure */
    pObject = (LPCONFIG_OBJECT) calloc(1, sizeof(CONFIG_OBJECT));
    if (pObject == NULL) {
        rc = ERR_MEM_ALLOC_FAILED;
        goto end_of_function;
    }

    /* init */
    pObject->physical_id        = physical_id;
    pObject->pszName            = pszName;
    pObject->type               = type;
    pObject->nModuleControlAsID = 0;
    pObject->pControllingModule = NULL;
    pObject->pReportIDList      = NULL;
    pObject->pReportModulesList = NULL;

    /* Vector inits */
    rc = vectorCreate(&(pObject->pReportIDList));
    if (rc) { goto end_of_function; }

    rc = vectorCreate(&(pObject->pReportModulesList));
    if (rc) { goto end_of_function; }

    /* Return to the caller */
    *ppObject = pObject;

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return (rc);
}

int configObjectGetId(LPCONFIG_OBJECT pObject, unsigned int * pnId)
{
    int rc = 0;

    *pnId = pObject->physical_id;

    return rc;
}

int configObjectGetName(LPCONFIG_OBJECT pObject, char ** pszName)
{
    int rc = 0;

    if (pObject->pszName != NULL) {
        *pszName = strdup(pObject->pszName);
    }
    else {
        *pszName = NULL;
    }

    return rc;
}

int configObjectGetType(LPCONFIG_OBJECT pObject, enum displayObjectType * pType)
{
    int rc = 0;

    *pType = pObject->type;

    return rc;
}



