/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <string.h>

#if !defined(_WIN32)
#   include <time.h>
#   include <unistd.h>
#   include <termios.h>
#   include <sys/select.h>
#   include <sys/time.h>
#   include <pthread.h>
#   include <dlfcn.h>
#endif

#include "lcd.h"
#include "trace.h"
#include "display.h"
#include "eventQueue.h"
#include "module.h"
#include "vector.h"
#include "driver.h"
#include "module2.h"

/*
 * This file is the "display" module.
 */
#define IILC_MODULE MODULE_TYPE_DISPLAY


/*
 * Macros to convert fan speeds to and from an int value
 */

#define FAN_SPEED_TO_INT(fan,percent)   (((fan & 0xFFFF) << 16) + (percent & 0x0000FFFF))
#define FAN_ID_FROM_INT(val)            ((val & 0xFFFF0000) >> 16)
#define FAN_SPEED_FROM_INT(val)         (val & 0x0000FFFF)



/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

#define DRIVER_PRE_INVOKE(pDev, pfn) \
    ((pDev == NULL) ? ERR_BAD_PARAMETER \
        : ((pDev->pFnArray->pfn == NULL) ? ERR_NOT_IMPLEMENTED \
            : pthread_mutex_lock(&(pDev->mtxModule))))

#define DRIVER_POST_INVOKE(pDev) pthread_mutex_unlock(&(pDev->mtxModule))

/* Each platform has its own way of dealing with dynamic libraries */
#if defined(_WIN32)
#   define LIB_NAME_PREFIX
#   define LIB_NAME_SUFFIX         ".dll"
#   define DYNAMIC_LIB_HANDLE      HMODULE
#else
#   define LIB_NAME_PREFIX         "lib"
#   define LIB_NAME_SUFFIX         ".so"
#   define DYNAMIC_LIB_HANDLE      void *
#endif



typedef struct _tag_led_instance
{
    /** Colour of the LED when it is "on" */
    COLOUR on;

    /** Colour of the LED when it is "off" */
    COLOUR off;

    /** The percentage of time which the LED spends on */
    unsigned int dutyCycle;

    /** The frequency of cycling from off->on->off in Hz/100 */
    unsigned int freq;

    /** Internal flag to track if the LED is on or off */
    int bLEDOn;

    /** The callback function used to decide this LED's status */
    //pfnGetLEDStatus pfn;

    /** The context to supply to the above callback function */
    void * context;

    /** The ID presented to the module controlling this LED */
    int led_id;

} LED_INSTANCE, * LPLED_INSTANCE;


typedef struct _tag_display_set_fan_power
{
    unsigned int fan_id;

    unsigned int fan_power;

} FAN_POWER;

typedef struct _tag_display_object_listener
{
    /** The module to receive the callback information */
    LPMODULE pModule;

    /** ID of the device object which this relates to */
    unsigned int id;

    /** Function to receive report information */
    union
    {
        pfnDspCBFanSpeed FanSpeed;
        pfnDspCBKeyEvent KeyEvent;
        pfnDspCBTemp     Temp;
    }
    pfn;

    /** Context data to supply to the function */
    void * context;

    /** Human-readable name */
    const char * pszName;
}
OBJECT_LISTENER, * LP_OBJECT_LISTENER;

typedef struct _tag_display_temp_listener
{
    /** Sensor ID which this relates to */
    unsigned int sensor_id;

    /** Function to receive temperature report */
    pfnDrvCBTemp pfn;

    /** Context data to supply to the function */
    void * context;

    /** Human-readable name of the sensor */
    const char * pszFanName;
} 
_TEMP_LISTENER, * LP_TEMP_LISTENER;

typedef struct _tag_display_fan_listener
{
    /** Fan ID which this relates to */
    unsigned int fan_id;

    /** Function to receive fan speed report */
    pfnDspCBFanSpeed pfn;

    /** Context data to supply to the function */
    void * context;

    /** Human-readable name of the fan */
    const char * pszFanName;
}
FAN_LISTENER, * LPFAN_LISTENER;


typedef struct _tag_display
{
    /** Platform-specific handle to the dynamic module */
    DYNAMIC_LIB_HANDLE libraryHandle;

    /** Context data to supply to the module when we call it */
    LP_DRIVER_INSTANCE pDriver;

    /** Port which this device is connected to */
    char * pszPort;

    pthread_mutex_t mtxModule;

    struct
    {
        pfnDrvCBKeyEvent  KeyEvent;

        pfnDrvCBTemp TempReport;

        pfnDrvCBFanSpeed    FanSpeedReport;

    } pfn;

    /** An array of functions which the module uses to indicate capabilities */
    LP_DRIVER_FN_ARRAY pFnArray;

    /** A private array of functions that are really the implementation ptrs */
    LP_DRIVER_FN_ARRAY pPrivateFnArray;

    struct
    {
        void * FanSpeedReport;

        void * KeyEvent;

        void * TempReport;
    
    } context;

    /* ----- Physical characteristics of the display ----- */

    /* Physical width of the device */
    unsigned int width;

    /* Physical height of the device */
    unsigned int height;

    /* Number of LEDs physically available on this display */
    unsigned int nLED;

    /* Number of fans physically available on this display */
    unsigned int nFans;

    /** Number of temperature sensors connected */
    unsigned int nTempSensors;

    /* ----- LEDs ----- */

    /** List of data regions relating to each LED */
    LPLED_INSTANCE * pLEDInstanceList;

    /* ----- Fans ----- */

    LPVECTOR pFanListeners;

    /* ----- Temperature sensors ----- */

    LPVECTOR pTempListeners;

    /* ----- LEDs ----- */

    pfnDrvCBKeyEvent pfnKeyListener;

    void * pKeyListenerCtxt;


    /* ----- Screen information ----- */

    /** The current screen being displayed */
    int nScreenCurrent;

    /** The list of screens to be displayed */
    LPVECTOR pScreenList;

    /** List of threads which control the rendering of each screen */
    pthread_t * screenThreads;

    /* ----- Others ----- */

    /* The event queue around which the whole Tx loop rotates */
    LPEVENT_QUEUE    pEventQueue;

    /* ID of the Txthread */
    pthread_t        TxThreadID;
}
DISPLAY;


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/


void * TxThread(void * arg);

void * RenderThread(void * arg);


int displayRefreshScreen(LPDISPLAY pDisplay);

/*

void displayCallbackKeyEvent(void * context, PACKET_KEY_EVENT event);
*/

void displayCallbackFan(void * context, unsigned int fan, unsigned int rpm);

void displayCallbackTemp(void * context, unsigned int sensor, double degc);

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/



/**
  * \brief Creates an instance of the display.
  *
  * In terms of the configuration file, this is representative of a "display"
  * object. 
  *
  * \param pszClass The class of driver to use when establishing communications
  *                 with the physical device. This forms the name of the plugin
  *                 library to load. This value is non-NULL and is a NUL-terminated
  *                 string.
  * \param pszPort  A free-form string field which is meaningful to the driver
  *                 in use. This value is non-NULL and is a NUL-terminated string.
  *                 For the exact values allowed in this string, see the documentation
  *                 for the display driver.
  * \param ppDisplay Non-NULL to receive the pointer to the new display object.
  *
  * \retval int 0 on success. All other values indicate failure.
  */
int displayCreate(const char * pszClass, const char * pszPort, LPDISPLAY * ppDisplay)
{
    int rc = 0;

    LPDISPLAY pDisplay = NULL;

    pfnDriverCreate pfnCreate = NULL;

    char libraryName[512];
    char * dlError = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayCreate);
    traceDebug(IILC_TRACE_CONTEXT,
        "pszClass %p, pszPort %p, ppDisplay %p", pszClass, pszPort, ppDisplay);

    if (pszClass == NULL || pszPort == NULL || ppDisplay == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT,
        "Creating display of type [%s] on port [%s]", pszClass, pszPort);

    /* Allocate the main memory */
    pDisplay = (LPDISPLAY) calloc(1, sizeof(DISPLAY));

    /* Take a copy of the port string */
    pDisplay->pszPort = (char *) calloc(strlen(pszPort)+1, sizeof(char));
    strcpy(pDisplay->pszPort, pszPort);

    /* Create the full module name */
    sprintf(libraryName, LIB_NAME_PREFIX "%s" LIB_NAME_SUFFIX, pszClass);

    /* Dynamically load the library */
#if defined(_WIN32)
    pDisplay->libraryHandle = LoadLibrary(libraryName);
#else
    pDisplay->libraryHandle = dlopen(libraryName, RTLD_NOW);
    dlError = dlerror();
#endif

    /* Verify load worked */
    if (pDisplay->libraryHandle == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Library load failed: %s", dlError);
        rc = ERR_INVALID_CONFIG;
        goto end_of_function;
    }

    /* Find the named entry point into the module */
#if defined(_WIN32)
#else
    dlerror();
    pfnCreate = (pfnDriverCreate) dlsym(pDisplay->libraryHandle, "driver_Create");
    dlError = dlerror();
#endif

    if (pfnCreate == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Failed loading module create function: %s", dlError);
        goto end_of_function;
    }

    /* Create a region of memory to hold the function pointers */
    pDisplay->pFnArray = (LP_DRIVER_FN_ARRAY) calloc(1, sizeof(DRIVER_FN_ARRAY));
    
    eventQueueCreate(&(pDisplay->pEventQueue));

    vectorCreate(&(pDisplay->pFanListeners));
    vectorCreate(&(pDisplay->pTempListeners));
    vectorCreate(&(pDisplay->pScreenList));

    /* Now call the module init routine */
    rc = (pfnCreate)(pDisplay->pszPort, pDisplay->pFnArray, &(pDisplay->pDriver));
    if (rc) {
        goto end_of_function;
    }


    /* Let it know which routine to use for fan callbacks */
    (pDisplay->pFnArray->SetCallbackFan)(pDisplay->pDriver, displayCallbackFan, pDisplay);
    (pDisplay->pFnArray->SetCallbackTemp)(pDisplay->pDriver, displayCallbackTemp, pDisplay);

    pDisplay->height = 4;
    pDisplay->width = 20;

    /* Assign to user area */
    *ppDisplay = pDisplay;
    
end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}




int displayDispose(LPDISPLAY * ppDisplay)
{
    int rc = 0;
    LPDISPLAY pDev = NULL;

    if (ppDisplay == NULL) {
        goto end_of_function;
    }

    pDev = *ppDisplay;

    if (pDev->pszPort != NULL) {
        free(pDev->pszPort);
    }

    free(pDev);

end_of_function:

    return rc;
}





int displayStart(LPDISPLAY pDisplay)
{
    int rc = 0;
    EVENT evt;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayStart);
    traceDebug(IILC_TRACE_CONTEXT, "pDisplay %p", pDisplay);

    /* Start the underlying device */
    (pDisplay->pFnArray->Start)(pDisplay->pDriver);

    /* Spin off a transmit thread */
    rc = pthread_create(&(pDisplay->TxThreadID), NULL, TxThread, (void *) pDisplay);
    if (rc) {
        printf("pthread_create error: %s\n", strerror(errno));
        goto end_of_function;
    }

    /*
     * Enqueue some default actions
     */

    /* Create a display clear event */
    EVENT_INIT(evt, EVENT_SCREEN_CLEAR, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Poll each module for data */
    EVENT_INIT(evt, EVENT_POLL_MODULE, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Contrast to something useful */
    EVENT_INIT(evt, EVENT_SET_CONTRAST, NULL, 90);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Set the backlight to 30% */
    EVENT_INIT(evt, EVENT_SET_BACKLIGHT, NULL, 30);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Start fans 0 & 1 at 0% */
    EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, FAN_SPEED_TO_INT(0, 50));
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, FAN_SPEED_TO_INT(1, 50));
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Update the screen regularly */
    EVENT_INIT(evt, EVENT_SCREEN_REFRESH, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Repeatedly poll module 0 */
    EVENT_INIT(evt, EVENT_POLL_MODULE, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);


end_of_function:

    IILC_TRACE_EXIT_RC(rc);
	return rc;
}


/**
  * Informs the display of the function to use when checking the fan power
  * level.
  *
  * \param pDisplay Display instance to which this relates.
  * \param fan_id The fan number as found in the display which is to be controlled.
  * \param pfn Pointer to a function which will receive the fan power request.
  * Using NULL here stops the fan being controlled by any external logic.
  * \param context Context data to provide to the callback function. May be NULL.
  *
  * \retval Zero on success. Any other value indicates error.
  */



int displayStop(LPDISPLAY pDisplay)
{
	int rc = 0;
    int i  = 0;
    //int ret;
    EVENT evt;
    LPEVENT_QUEUE pEventQueue = pDisplay->pEventQueue;

    /* ----- Restore display to sensible state ----- */

    /* Switch fans back on */
    EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, FAN_SPEED_TO_INT(0, 100));
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    EVENT_INIT(evt, EVENT_SET_FAN_STATE, NULL, FAN_SPEED_TO_INT(1, 100));
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    /* Stop fan reporting for all fans */
    for (i = 0; i < 4; i++) {
        EVENT_INIT(evt, EVENT_DISABLE_FAN_REPORT, NULL, i);
        eventQueueEnqueue(pEventQueue, &evt, 0);
    }

    /* Backlight off */
    EVENT_INIT(evt, EVENT_SET_BACKLIGHT, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);


    /* Terminate the Tx thread */
    EVENT_INIT(evt, EVENT_SHUTDOWN, NULL, 0);
    eventQueueEnqueue(pEventQueue, &evt, 0);

    /* Wait for the Tx thread to complete */
    pthread_join(pDisplay->TxThreadID, NULL);

    /* Shutdown the device itself */
    (pDisplay->pFnArray->Stop)(pDisplay->pDriver);

    /* Release this memory */
//    free(pDisplay);

	/* Clear user's reference */
//	pDisplay = NULL;

//end_of_function:

	return rc;
}

typedef enum _tag_window_type
{
    WINDOW_TYPE_TEXT,
    WINDOW_TYPE_SCROLL_TEXT,
    WINDOW_TYPE_WRAP_TEXT,
    WINDOW_TYPE_BAR,
    WINDOW_TYPE_LARGE_TEXT

} WINDOW_TYPE;




/**
 * Registers a callback which will receive regular reports regarding
 * the speed of fans connected to the display.
 *
 * \param pDisplay Pointer to the display which is being modified.
 * \param fan_id The physical fan number connected to the display.
 * \param pfnListener Callback function to receive the fan reports.
 * \param context Context data to be supplied to the callback function.
 * \param report_as_id A value to supply to the callback as a way of
 *        abstracting the physical fan ID from the fan ID received by
 *        the callback.
 *
 * \retval Zero on success. All other values indicate error.
 */

int displayAddFanReportListener(LPDISPLAY pDisplay, unsigned int fanId, LPMODULE pModule, const char * pszName)
{
    int rc = 0;
    EVENT evt;
    LP_OBJECT_LISTENER pListener = NULL;

    /* Set the new data */
    pListener = (LP_OBJECT_LISTENER) malloc(sizeof(OBJECT_LISTENER));
    pListener->pModule      = pModule;
    pListener->id           = fanId;
    pListener->pszName      = pszName;
    vectorAdd(pDisplay->pFanListeners, pListener);

    /* Add an update for this fan to the event queue */
    EVENT_INIT(evt, EVENT_ENABLE_FAN_REPORT, NULL, fanId);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

//end_of_function:

    return rc;
}

int displayRemoveFanReportListener(LPDISPLAY pDisplay, LPMODULE pModule)
{
    return 0;
}

int displayAddTempReportListener(LPDISPLAY pDisplay, unsigned int sensor_id, LPMODULE pModule, const char * pszReportAs)
{
    int rc = 0;
    EVENT evt;
    LP_OBJECT_LISTENER pListener = NULL;

    /* Set the new data */
    pListener = (LP_OBJECT_LISTENER) malloc(sizeof(OBJECT_LISTENER));
    pListener->pModule   = pModule;
    pListener->id        = sensor_id;
    pListener->pszName   = pszReportAs == NULL ? NULL : strdup(pszReportAs);
    vectorAdd(pDisplay->pTempListeners, pListener);

    /* Add an update for this temp to the event queue */
    EVENT_INIT(evt, EVENT_ENABLE_TEMP_REPORT, NULL, sensor_id);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

//end_of_function:

    return rc;
}

int displayRemoveTempReportListener(LPDISPLAY pDisplay, pfnDspCBTemp pfnListener)
{
    return 0;
}

int displayAddKeyEventListener(LPDISPLAY pDisplay, pfnDspCBKeyEvent pfnListener, void * context)
{

    return 0;
}



/**
  * Main thread function which is used for updating the display.
  */
void * TxThread(void * arg)
{
    /* Grab our instance data from the arg ptr */
    LPDISPLAY pDisplay = (LPDISPLAY) arg;
    
    /* Vars used in this function */
    int running = 1;
    int rc = 0;

    /* Initialise convenient locals */
    LPEVENT_QUEUE pEventQueue = pDisplay->pEventQueue;

    /* Temp */
    char * longText = "This is a test";
    //char scrData[1024];
    int scrollPos = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_TxThread);

    while (running)
    {
        /* Loop vars */
        EVENT evt;

        /* Flag to indicate we should keep the memory for the event (saves reallocing) */
        int iRequeueTime = -1; /* -1 == dispose of event */

        /* Wait for event dequeue */
        rc = eventQueueDequeue(pEventQueue, &evt);
        if (rc) {
            traceError(IILC_TRACE_CONTEXT, "Event dequeue failed");
            break;
        }

        traceDebug(IILC_TRACE_CONTEXT, "Dequeued event %d", evt.eType);

        /* Process */
        switch (evt.eType)
        {
        case EVENT_SHUTDOWN:
            running = 0;
            break;

        case EVENT_SCREEN_CLEAR:
//            deviceClear(pDisplay->pDevice);
            break;

        case EVENT_SET_BACKLIGHT:
            (pDisplay->pFnArray->SetBacklight)(pDisplay->pDriver, (unsigned int) evt.iData);
            break;

        case EVENT_SET_CONTRAST:
            (pDisplay->pFnArray->SetContrast)(pDisplay->pDriver, (unsigned char) evt.iData);
            break;

        case EVENT_SCREEN_CHANGE:
//            rc = displayAdvanceScreen(pDisplay);
            iRequeueTime = 5000;
            break;

        case EVENT_SCREEN_REFRESH:
            rc = displayRefreshScreen(pDisplay);
            iRequeueTime = 500;
            break;

        case EVENT_LED_UPDATE:
        {
            /* Convenience */
            LPLED_INSTANCE pLEDInst = pDisplay->pLEDInstanceList[evt.iData];

            printf("Updating LED %d status\n", evt.iData);

            /* Get current state and adjust output accordingly */
            if (pLEDInst->bLEDOn || pLEDInst->freq == 0) {
                (pDisplay->pFnArray->SetLED)(pDisplay->pDriver, evt.iData,
                    pLEDInst->off.red, pLEDInst->off.green, pLEDInst->off.blue);
            }
            else {
                (pDisplay->pFnArray->SetLED)(pDisplay->pDriver, evt.iData, 
                    pLEDInst->on.red, pLEDInst->on.green, pLEDInst->on.blue);
            }

            /* Calculate the requeue time */
            if (pLEDInst->freq == 0) {
                iRequeueTime = 2000; /* Frequency of zero means always off */
            }
            else {
                iRequeueTime = pLEDInst->bLEDOn ?
                    ((1000 * pLEDInst->dutyCycle) / pLEDInst->freq) :
                    ((1000 * (100 - pLEDInst->dutyCycle)) / pLEDInst->freq);
            }

            /* Toggle state and loop */
            pLEDInst->bLEDOn = ~(pLEDInst->bLEDOn);
            break;
        }

        /*
        case EVENT_POLL_LED_STATUS:
        {
            LPLED_INSTANCE pLEDInst = pDisplay->pLEDInstanceList[evt.iData];

            printf("Polling for LED %d status\n", evt.iData);

            * Get LED data from module *
            (pLEDInst->pfn)(pLEDInst->context, pLEDInst->led_id,
                &(pLEDInst->on), &(pLEDInst->off),
                &(pLEDInst->dutyCycle), &(pLEDInst->freq));

            * LEDs need only be updated every 2 seconds *
            iRequeueTime = 2000;
            break;

        }
        */

        case EVENT_POLL_MODULE:
        {
            break;
        }

        case EVENT_TEXT_SCROLL:
        {
            char tmpBuf[1024];
            size_t strLen = -1;
            int dispWidth = 20;
            
            /* Amount of string to end */
            strLen = strlen(&longText[scrollPos]);

            if (strLen > (unsigned int) dispWidth) {
                sprintf(tmpBuf, "%.*s", dispWidth, &longText[scrollPos]);
            }
            else {
                sprintf(tmpBuf, "%s %.*s", &longText[scrollPos], dispWidth - strLen - 1, longText);
            }

            /* Write to the display */
            (pDisplay->pFnArray->WriteText)(pDisplay->pDriver, 0, 1, tmpBuf, (unsigned int) strlen(tmpBuf));

            /* Scroll and check within bounds */
            scrollPos++;
            if ((unsigned int) scrollPos >= strlen(longText)) {
                scrollPos = 0;
            }

            /* Keep event in memory and requeue for 50 msec time */
            iRequeueTime = 300;
            break;
        }

        case EVENT_KEY_PRESS:
            break;

        case EVENT_KEY_RELEASE:
            break;

        case EVENT_ENABLE_FAN_REPORT:
            (pDisplay->pFnArray->EnableFanReport)(pDisplay->pDriver, evt.iData);
            traceDebug(IILC_TRACE_CONTEXT, "Enabled fan reporting for %d", evt.iData);
            break;

        case EVENT_DISABLE_FAN_REPORT:
            (pDisplay->pFnArray->DisableFanReport)(pDisplay->pDriver, evt.iData);
            traceDebug(IILC_TRACE_CONTEXT, "Disabled fan reporting for %d", evt.iData);
            break;

        case EVENT_ENABLE_TEMP_REPORT:
            (pDisplay->pFnArray->EnableTempReport)(pDisplay->pDriver, evt.iData);
            traceDebug(IILC_TRACE_CONTEXT, "Enabled temp reporting for %d", evt.iData);
            break;

        case EVENT_DISABLE_TEMP_REPORT:
            (pDisplay->pFnArray->DisableTempReport)(pDisplay->pDriver, evt.iData);
            traceDebug(IILC_TRACE_CONTEXT, "Disabled temp reporting for %d", evt.iData);
            break;

        case EVENT_SET_FAN_STATE:
            (pDisplay->pFnArray->SetFanPower)(pDisplay->pDriver, 
                FAN_ID_FROM_INT(evt.iData), FAN_SPEED_FROM_INT(evt.iData));
            iRequeueTime = 3000;
            break;

        default:
            printf("Unknown event type received (%d)\n", evt.eType);
            break;
        }

        /* Should we requeue this event? */
        if (iRequeueTime >= 0) {
            eventQueueEnqueue(pEventQueue, &evt, iRequeueTime);
        }
    }


//end_of_function:

    IILC_TRACE_EXIT;
    return NULL;
}


int displayGetSize(LPDISPLAY pDisplay, int * width, int * height)
{
    int rc = 0;

    *width = pDisplay->width;
    *height = pDisplay->height;

    return rc;
}

int displayScreenAdd(LPDISPLAY pDisplay, LPSCREEN pScreen)
{
    int rc = 0;
    //LPSCREEN * pNewList = NULL;
    //pthread_t * pNewThreadList = NULL;
    //pthread_t newTid;
    EVENT evt;

    vectorAdd(pDisplay->pScreenList, pScreen);

    /* Remember to update us in the main Tx loop */
    EVENT_INIT(evt, EVENT_SCREEN_REFRESH, NULL, 0);
    eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

    return rc;
}



int displayRefreshScreen(LPDISPLAY pDisplay)
{
    int rc = 0;
    unsigned int i = 0;
    //int nWindows = 0;
    LPSCREEN pScreen = NULL;
    int nScreenCount = 0;
    
    /* Buffer to draw to screen */
    char * pszScreen = NULL;
    char * ptrCurrent = NULL;

    /* Allocate some (2-D) memory */
    pszScreen = (char *) calloc(pDisplay->width * pDisplay->height, sizeof(char));

    /* If we have no screens, let the user know */
    vectorGetSize(pDisplay->pScreenList, &nScreenCount);
    if (nScreenCount == 0) {
        char * pszNoConfig = "No screens setup";
        strncpy(&pszScreen[0], pszNoConfig, strlen(pszNoConfig));
    }
    else {
        /* Get the current screen we are displaying */
        vectorGetElementAt(pDisplay->pScreenList, pDisplay->nScreenCurrent, (void **) &pScreen);

        /* Render this line into the current buffer */
        screenDoRender(pScreen, pszScreen);
    }

    /* Write all data to the screen */
    ptrCurrent = pszScreen;
    for (i = 0; i < pDisplay->height; i++) {
        rc = (pDisplay->pFnArray->WriteText)(pDisplay->pDriver, 0, i, ptrCurrent, pDisplay->width);
        ptrCurrent += pDisplay->width;
    }

    /* Release memory */
    if (pszScreen) free(pszScreen);

    return rc;
}


/**
  * Function registered with the device driver to receive temperature
  * information via callbacks. For full documentation see the definition
  * of the pfnDrvCBTemp function pointer found in driver.h.
  */
void displayCallbackTemp(void * context, unsigned int sensor, double degc)
{
    int i = 0;
    int nCount = 0;
    LPDISPLAY pDisplay = (LPDISPLAY) context;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayCallbackTemp);
    traceDebug(IILC_TRACE_CONTEXT, "sensor %u, temp %f", sensor, degc);

    /* Iterate through all callbacks */
    vectorGetSize(pDisplay->pTempListeners, &nCount);
    for (i = 0; i < nCount; i++) {

        LP_OBJECT_LISTENER pListener;

        /* Get the next callback */
        vectorGetElementAt(pDisplay->pTempListeners, i, (void **) &pListener);

        /* Is this callback relevant for this sensor? */
        if (pListener->id != sensor) {
            continue; /* No - skip to next callback */
        }

        /* Yes - perform the callback */
        moduleReportTemp(pListener->pModule, sensor, degc, pListener->pszName);
    }

    IILC_TRACE_EXIT;
}


void displayCallbackFan(void * context, unsigned int fan, unsigned int rpm)
{
    int i = 0;
    int nCount = 0;
    LPDISPLAY pDisplay = (LPDISPLAY) context;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displayCallbackFan);
    traceDebug(IILC_TRACE_CONTEXT, "fan %u, rpm %u", fan, rpm);

    /* Iterate through all callbacks */
    vectorGetSize(pDisplay->pFanListeners, &nCount);
    for (i = 0; i < nCount; i++) {

        LP_OBJECT_LISTENER pListener;

        /* Get the next callback */
        vectorGetElementAt(pDisplay->pFanListeners, i, (void **) &pListener);

        /* Is this callback relevant for this fan? */
        if (pListener->id != fan) {
            continue; /* No - skip to next callback */
        }

        /* Yes - perform the callback */
        moduleReportFan(pListener->pModule, fan, rpm, pListener->pszName);
    }

    IILC_TRACE_EXIT;
}


/**
  * Informs the display of the function to use when checking the fan power
  * level.
  *
  * \param pDisplay Display instance to which this relates.
  * \param led_id The LED number as found in the display which is to be controlled.
  * \param pfn Pointer to a function which will receive the LED display status.
  * Using NULL here stops the LED being controlled by any external logic, and will
  * switch the LED off.
  * \param context Context data to provide to the callback function. May be NULL.
  *
  * \retval Zero on success. Any other value indicates error.
  */
int displaySetLEDController(LPDISPLAY pDisplay, unsigned int led_id, void * /* pfnGetLEDStatus */ pfn, void * context, int report_as_led)
{
    int rc = 0;
    EVENT evt;
    LPLED_INSTANCE pInst = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_displaySetLEDController);
    traceDebug(IILC_TRACE_CONTEXT,
        "pDisplay " T_FMT_PTR ", led_id " T_FMT_INT ", pfn " T_FMT_PTR
        ", context " T_FMT_PTR ", report_as_led " T_FMT_INT,
        pDisplay, led_id, pfn, context, report_as_led);

    /* Sanity checks */
    if (pDisplay == NULL) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    if (led_id > pDisplay->nLED) {
        rc = ERR_BAD_PARAMETER;
        goto end_of_function;
    }

    /* Convenience */
    pInst = pDisplay->pLEDInstanceList[led_id];

    /* Clear and init memory */
    //pInst->     = pfn;
    pInst->context = context;
    pInst->led_id  = report_as_led;

    /* Are we setting up or removing a controller? */
    if (pfn != NULL) {

        /* Remember to query this LED */
        EVENT_INIT(evt, EVENT_POLL_LED_STATUS, NULL, led_id);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);

        /* Remember to update this LED */
        EVENT_INIT(evt, EVENT_LED_UPDATE, NULL, led_id);
        eventQueueEnqueue(pDisplay->pEventQueue, &evt, 0);
    }

end_of_function:

    IILC_TRACE_EXIT;
    return 0;
}
