/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <errno.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#   include "time_w32.h"
#else
#   include <sys/time.h>
#endif

#include "eventQueue.h"


typedef struct _tag_event_queue_entry * LPEVENT_QUEUE_ENTRY;

typedef struct _tag_event_queue_entry
{
    /* Queue data */
    EVENT eventData;

    /* Previous event in this doubly-linked list */
    LPEVENT_QUEUE_ENTRY pPrevEvent;

    /* Next event in this doubly-linked list */
    LPEVENT_QUEUE_ENTRY pNextEvent;

    /* Time at which this event should dequeue */
    struct timeval dequeueTime;

} EVENT_QUEUE_ENTRY;


typedef struct _tag_event_queue
{
    /* Queue mutex */
    pthread_mutex_t mtxQueue;

    /* Item enqueued sempahore */
    pthread_cond_t semAvailable;

    /* Front item in the queue */
    LPEVENT_QUEUE_ENTRY pFirst;

    /* Last item in the queue */
    LPEVENT_QUEUE_ENTRY pLast;

} EVENT_QUEUE;



int eventQueueCreate(LPEVENT_QUEUE * ppEventQueue)
{
    int rc = 0;

    LPEVENT_QUEUE pEventQueue = NULL;

    pEventQueue = malloc(sizeof(EVENT_QUEUE));

    /* Create the mutex */
    rc = pthread_mutex_init(&(pEventQueue->mtxQueue), NULL);
    if (rc) {
        printf("Failed to create queue mutex: %s\n", strerror(errno));
        goto end_of_function;
    }
    
    /* Create the semaphore */
    rc = pthread_cond_init(&(pEventQueue->semAvailable), NULL);
    if (rc) {
        printf("Failed to create queue semaphore: %s\n", strerror(errno));
        goto end_of_function;
    }

    
    *ppEventQueue = pEventQueue;

end_of_function:

    return 0;
}

int eventQueueDispose(LPEVENT_QUEUE * ppEventQueue)
{
    LPEVENT_QUEUE pEventQueue = NULL;
    LPEVENT_QUEUE_ENTRY pCurrEntry = NULL;

    pEventQueue = *ppEventQueue;

    /* Lock the queue */
    pthread_mutex_lock(&(pEventQueue->mtxQueue));

    /* Walk across deleting everything */
    pCurrEntry = pEventQueue->pFirst;
    while (pCurrEntry != NULL) {
        free(pCurrEntry);
    }

    /* Clear the front & back */
    pEventQueue->pFirst = NULL;
    pEventQueue->pLast = NULL;

    /* Release the lock */
    pthread_mutex_unlock(&(pEventQueue->mtxQueue));

    pthread_cond_destroy(&(pEventQueue->semAvailable));

    pthread_mutex_destroy(&(pEventQueue->mtxQueue));

    free(pEventQueue);

    *ppEventQueue = NULL;

    return 0;
}


int eventQueueEnqueue(LPEVENT_QUEUE pEventQueue, LPEVENT pEvent, unsigned int fireInMillis)
{
    LPEVENT_QUEUE_ENTRY pNew = NULL;
    LPEVENT_QUEUE_ENTRY pCurr = NULL;
    //LPEVENT_QUEUE_ENTRY pNext = NULL;
    
    /* Create the entry (OUTSIDE the mutex lock) */
    pNew = malloc(sizeof(EVENT_QUEUE_ENTRY));
    pNew->pNextEvent = NULL;
    pNew->pPrevEvent = NULL;
    memcpy(&(pNew->eventData), pEvent, sizeof(EVENT));

    /* Set the time when this is to dequeue */
    gettimeofday(&(pNew->dequeueTime), NULL);
    
    /* Increment seconds */
    pNew->dequeueTime.tv_sec += fireInMillis / 1000;

    /* Increment milliseconds */
    pNew->dequeueTime.tv_usec += (fireInMillis % 1000) * 1000;

    /* Rationalise */
    if (pNew->dequeueTime.tv_usec > 1000000) {
        pNew->dequeueTime.tv_usec -= 1000000;
        pNew->dequeueTime.tv_sec  += 1;
    }

    /* Lock the queue mutex */
    pthread_mutex_lock(&(pEventQueue->mtxQueue));

    /* Get the item at the back of the queue */
    pCurr = pEventQueue->pLast;
    
    /* Now work forwards until we find one with the same or higher priority */
    while (pCurr != NULL) {

        struct timeval currDequeue = pCurr->dequeueTime;

        /* Decide if this to dequeue first */
        if ((currDequeue.tv_sec < pNew->dequeueTime.tv_sec) ||
            (currDequeue.tv_sec == pNew->dequeueTime.tv_sec && 
             currDequeue.tv_usec < pNew->dequeueTime.tv_usec)) {

            /* pCurr should dequeue before us - fall in line here */
            break;
        }

        /* Move forward in the queue */
        pCurr = pCurr->pPrevEvent;
    };

    /* Point the new item to the current one */
    pNew->pPrevEvent = pCurr;

    /* Are we at the front? */
    if (pCurr == NULL) {

        /* Point the new item to the original front item */
        pNew->pNextEvent = pEventQueue->pFirst;

        /* Point the queue front to new item */
        pEventQueue->pFirst = pNew;
    }
    else {
        /* Point the new one to the next one */
        pNew->pNextEvent = pCurr->pNextEvent;

        /* Point the current item to the new one */
        pCurr->pNextEvent = pNew;
    }

    /* Fix up the back of queue or the next queue item */
    if (pNew->pNextEvent == NULL) {
        pEventQueue->pLast = pNew; /* At back of queue */
    }
    else {
        pNew->pNextEvent->pPrevEvent = pNew;
    }

    /* Signal something in the queue has changed */
    pthread_cond_signal(&(pEventQueue->semAvailable));    

    /* Release the queue lock */
    pthread_mutex_unlock(&(pEventQueue->mtxQueue));

    return 0;
}


/**
  * Dequeues an event from the queue when it due.
  *
  * \param pEvent Pointer to an EVENT structure to be populated
  * with the event data
  */
int eventQueueDequeue(LPEVENT_QUEUE pEventQueue, LPEVENT pEvent)
{
    int bGotLock = 0;
    LPEVENT_QUEUE_ENTRY pEntry = NULL;
    struct timeval dequeueTime;
    struct timeval currentTime;
    
    /* Lock the queue mutex */
    pthread_mutex_lock(&(pEventQueue->mtxQueue));
    bGotLock = 1;

    /* Wait until something is available at front of queue */
    while (pEventQueue->pFirst == NULL) {
        pthread_cond_wait(&(pEventQueue->semAvailable), &(pEventQueue->mtxQueue));
    }

    /* Now something at front of queue and we have lock */
    pEntry = pEventQueue->pFirst;

    /* Get the current time */
    gettimeofday(&currentTime, NULL);

    /* When do we dequeue this? */
    dequeueTime = pEntry->dequeueTime;

    /* Can we dequeue yet? */
    if ((dequeueTime.tv_sec > currentTime.tv_sec) ||
        (dequeueTime.tv_sec == currentTime.tv_sec && dequeueTime.tv_usec > currentTime.tv_usec))  {

        struct timespec timeout;

        /* Can't dequeue yet.
         * We need to wait until the required time has elapsed or
         * something new joins the queue with a more urgent need.
         */

        timeout.tv_sec  = dequeueTime.tv_sec;
        timeout.tv_nsec = dequeueTime.tv_usec * 1000;

        /* Perform the wait */
        pthread_cond_timedwait(&(pEventQueue->semAvailable), &(pEventQueue->mtxQueue), &timeout);
        
        /* Now either we've timed-out, or something more urgent arrived.
         * In any case, dequeue immediately and return to user
         */
    }
    else {
        /* OK to dequeue immediately */
    }

    /* Grab the fired event (don't forgot front of queue may have changed) */
    pEntry = pEventQueue->pFirst;
    memcpy(pEvent, &(pEntry->eventData), sizeof(EVENT));

    /* Fix up the front of queue */
    pEventQueue->pFirst = pEntry->pNextEvent;
    
    /* Fix up the first item */
    if (pEventQueue->pFirst != NULL) {
        pEventQueue->pFirst->pPrevEvent = NULL;
    }
    else {
        /* Nothing left in the queue - fix up back */
        pEventQueue->pLast = NULL;
    }

    /* Release the queue lock */
    pthread_mutex_unlock(&(pEventQueue->mtxQueue));

    /* Release memory for this entry */
    free(pEntry);


    return 0;
}


