/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#else
#   include <pthread.h>
#   include <dlfcn.h>
#endif

#include "trace.h"
#include "module.h"

/* Each platform has its own way of dealing with dynamic libraries */
#if defined(_WIN32)
#   define LIB_NAME_PREFIX
#   define LIB_NAME_SUFFIX ".dll"
#else
#   define LIB_NAME_PREFIX "lib"
#   define LIB_NAME_SUFFIX ".so"
#endif


#define IILC_MODULE MODULE_TYPE_APP


/* ------------------------------------------------------------------------ *
 *                                                                          *
 * Module creation / destruction routines                                   *
 *                                                                          *
 * ------------------------------------------------------------------------ */




int moduleInit(const char * moduleName, LPMODULE * ppModule)
{
    int rc = 0;
    char libraryName[512];
    char * dlError = NULL;

    LPMODULE          pModule  = NULL;
    LPMODULE_FN_ARRAY pFnArray = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_moduleInit);
    traceDebug(IILC_TRACE_CONTEXT, "moduleName %p, ppModule %p", moduleName, ppModule);

    if (ppModule == NULL) {
        traceError(IILC_TRACE_CONTEXT, "ppModule == NULL");
        goto end_of_function;
    }

    traceDebug(IILC_TRACE_CONTEXT, "Initialising module name %s", moduleName);

    /* Create main module memory */
    pModule = malloc(sizeof(MODULE));
    if (pModule == NULL) {
        goto end_of_function;
    }

    /* Initialise the module data */
    pModule->libraryHandle = NULL;
    pModule->moduleContext = NULL;
    pModule->pfnCreate     = NULL;
    pModule->pfnDestroy    = NULL;
    pModule->pFnArray      = NULL;

    /* All the other functions */
    pFnArray = calloc(1, sizeof(MODULE_FN_ARRAY));
    
    /* A mutex used to serialise comms with the module */
    rc = pthread_mutex_init(&(pModule->mtxModule), NULL);
    if (rc) {
        traceError(IILC_TRACE_CONTEXT, "Error creating module mutex: %s", strerror(errno));
        goto end_of_function;
    }

    /* Create the full module name */
    sprintf(libraryName, LIB_NAME_PREFIX "%s" LIB_NAME_SUFFIX, moduleName);

    /* Dynamically load the library */
#if defined(_WIN32)
    pModule->libraryHandle = LoadLibrary(libraryName);
#else
    pModule->libraryHandle = dlopen(libraryName, RTLD_NOW);
    dlError = dlerror();
#endif

    if (pModule->libraryHandle == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Library load failed: %s", dlError);
        rc = ERR_INVALID_CONFIG;
        goto end_of_function;
    }

#if defined(_WIN32)
#else
    dlerror();
    pModule->pfnCreate = (pfnModuleCreate) dlsym(pModule->libraryHandle, "ModuleCreate");
    dlError = dlerror();
#endif

    if (pModule->pfnCreate == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Failed loading module create function: %s", dlError);
        goto end_of_function;
    }

#if defined(_WIN32)
#else
    dlerror();
    pModule->pfnDestroy = (pfnModuleDestroy) dlsym(pModule->libraryHandle, "ModuleDestroy");
    dlError = dlerror();
#endif

    if (pModule->pfnDestroy == NULL) {
        traceError(IILC_TRACE_CONTEXT, "Failed loading module destroy function: %s", dlError);
        goto end_of_function;
    }

    /* Assign local var into the main module ptr */
    pModule->pFnArray = pFnArray;
    
    /* Now call the module init routine */
    (pModule->pfnCreate)(1, &(pModule->moduleContext), pModule->pFnArray);

    *ppModule = pModule;

end_of_function:

    /* Tear down if things go wrong */
    if (rc) {


    }

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


int moduleUnload(LPMODULE * ppModule)
{
    LPMODULE pModule = *ppModule;

    /* Call destroy on the module first */
    if (pModule->pfnDestroy != NULL) {
        (pModule->pfnDestroy)(pModule->moduleContext);
    }

    /* Delete the mutex */
    pthread_mutex_destroy(&(pModule->mtxModule));

    /* Release the function array memory */
    free(pModule->pFnArray);

    /* Unload the library */
#if defined(_WIN32)
    FreeLibrary(pModule->libraryHandle);
#else
    dlclose(pModule->libraryHandle);
#endif

    /* Release the module structure memory */
    free(pModule);

    *ppModule = NULL;

    return 0;
}

int moduleReportTemp(LPMODULE pModule, unsigned int sensor_id, double degrees_c, const char * pszName)
{
    int rc = 0;

    if (pModule != NULL && pModule->pFnArray->pfnTempReport != NULL) {
        (pModule->pFnArray->pfnTempReport)(pModule->moduleContext, sensor_id, degrees_c, pszName);
    }

    return rc;
}


int moduleReportFan(LPMODULE pModule, unsigned int fan_id, unsigned int rpm, const char * pszName)
{
    int rc = 0;

    if (pModule != NULL && pModule->pFnArray->pfnFanSpeedReport != NULL) {
        (pModule->pFnArray->pfnFanSpeedReport)(pModule->moduleContext, fan_id, rpm, pszName);
    }

    return rc;
}


/*
int moduleGetDisplayData(LPMODULE pModule, LPSCREEN pScreen, unsigned int row, char ** buf)
{
    int rc = 0;

    if (pModule == NULL) {
        rc = -1;
        goto end_of_function;
    }

    * Lock the module *

    * Invoke the function *
    if (pModule->pFnArray->pfnTextData != NULL) {
        //rc = (pModule->pFnArray->pfnTextData)(pModule->moduleContext, pScreen, row, buf);
    }
    else {
        *buf[0] = '\0';
        rc = -1;
    }

    * Unlock the module *


end_of_function:

    return rc;
}
*/


int moduleGetLEDStatus(LPMODULE pModule, unsigned int led_id,
                       LPCOLOUR pColourOn, LPCOLOUR pColourOff,
                       unsigned int * dutyCycle, unsigned int * freq)
{
    int rc = 0;

    if (pModule == NULL) {
        rc = -1;
        goto end_of_function;
    }

    if (pModule->pFnArray->pfnReadLEDStatus == NULL) {
        rc = -1;
        goto end_of_function;
    }

    /* Lock the module */

    /* Invoke the function */
    //rc = (pModule->pFnArray->pfnReadLEDStatus)(pModule->moduleContext,
      //  led_id, pColourOn, pColourOff, dutyCycle, freq);

    /* Unlock the module */


end_of_function:

    return rc;
}



int moduleGetFanPower(LPMODULE pModule, int fan_id, unsigned int * pFanPower)
{
    int rc = 0;

    if (pModule == NULL) {
        rc = -1;
        goto end_of_function;
    }

    /* Lock the module */
    pthread_mutex_lock(&(pModule->mtxModule));

    /* Invoke the function */
    //rc = (pModule->pFnArray->pfnReadFanPower)(pModule->moduleContext, fan_id, pFanPower);

    /* Unlock the module */
    pthread_mutex_unlock(&(pModule->mtxModule));


end_of_function:

    return rc;
}


int moduleRender(LPMODULE pModule, int nRow, int nWidth, char * pBuffer)
{
    int rc = 0;

    if (pModule->pFnArray->pfnTextData != NULL) {
        (pModule->pFnArray->pfnTextData)(pModule->moduleContext, nRow, nWidth, pBuffer);
    }
    else {
        sprintf(pBuffer, "[NO FN]%.*s", nWidth - 7, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
    }

    return rc;
}
