/*
 * Copyright Ian Burnett 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include "trace.h"
#include "parserImpl.h"

#define IILC_MODULE MODULE_TYPE_APP



/****************************************************************************
 *                                                                          *
 * Parser-related goodness                                                  *
 *                                                                          *
 ****************************************************************************/

/*
 * A flag to set to switch on flex debug mode
 */
extern int yydebug;

/*
 * The file read by the parser
 */
extern FILE * yyin;

/*
 * Key under which to find the current parser in TLS
 */
static pthread_key_t g_keyParserThis;

/**
  * Variable which holds ensures parser thread data is initialised once only.
  */
static pthread_once_t parse_init = PTHREAD_ONCE_INIT;

/**
  * Macro which ensures the parser is initialised once and once only
  */
#define PARSER_INITIALISE  (void) pthread_once(&parse_init, parserInitialise);


/****************************************************************************
 *                                                                          *
 * Internal structure definitions                                           *
 *                                                                          *
 ****************************************************************************/


/** \brief Structure used to hold current context for the parser currently
  *        executing on this thread.
  */
typedef struct _tag_parser_context
{
    LPCONFIG_FILE pFile;

    LPCONFIG_DISPLAY pDisplay;

    LPCONFIG_DEVICE pDevice;

    LPCONFIG_MODULE pModule;

    LPCONFIG_OBJECT pObject;

    LPCONFIG_SCREEN pScreen;
}
PARSER_CONTEXT, * LP_PARSER_CONTEXT;




/****************************************************************************
 *                                                                          *
 * Parse-time functions                                                     *
 *                                                                          *
 ****************************************************************************/


void parserInitialise(void)
{
    /* Initialise the stack depth TLS key */
    (void) pthread_key_create(&g_keyParserThis, NULL);
}

LP_PARSER_CONTEXT parserGetContext()
{
    LP_PARSER_CONTEXT pContext = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserGetCurrentConfig);

    PARSER_INITIALISE;

    /* Grab the current parser on this thread */
    pContext = (LP_PARSER_CONTEXT) pthread_getspecific(g_keyParserThis);
    if (pContext == NULL) {
        printf("Failed to obtain parser on current thread\n");
        exit(1);
    }

    traceDebug(IILC_TRACE_CONTEXT, "pContext %p", pContext);
    IILC_TRACE_EXIT;
    return pContext;
}


int parserParse(LPCONFIG_FILE * ppFile, const char * pszFileName)
{
    int rc = 0;
    LP_PARSER_CONTEXT pContext = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_parserParse);
    traceDebug(IILC_TRACE_CONTEXT,
        "pszConfigFile %p, ppFile %p", pszFileName, ppFile);

    PARSER_INITIALISE;

    /* Create a context block */
    pContext = (LP_PARSER_CONTEXT) calloc(1, sizeof(PARSER_CONTEXT));

    /* Create a config file instance */
    configFileCreate(&(pContext->pFile), pszFileName);

    /* Make this available on the current thread */
    pthread_setspecific(g_keyParserThis, (void *) pContext);

    /* Open the file */
    yyin = fopen(pszFileName, "r");
    yydebug = 0;

    /* Parse the file */
    yyparse();

    /* Close the file */
    fclose(yyin);

    /* Return config to user */
    *ppFile = pContext->pFile;

    /* Release memory */
    free(pContext);

    /* Drop the TLS data */
    pthread_setspecific(g_keyParserThis, NULL);

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}



/* ------------------------------------------------------------------------ */
/*                                                                          */
/* Functions used by the parser to read in the configuration                */
/*                                                                          */
/* ------------------------------------------------------------------------ */

/* ----------------- Device entity ----------------- */

void parserDeviceBegin(char * pszDevice)
{
    configDeviceCreate(&(parserGetContext()->pDevice), pszDevice);
}

void parserDeviceSetClass(char * pszClass)
{
    configDeviceSetClass(parserGetContext()->pDevice, pszClass);
}

void parserDeviceSetPort(char * pszPort)
{
    configDeviceSetPort(parserGetContext()->pDevice, pszPort);
}

int  parserDeviceObjectCreate(int object_id, char * pszName, enum displayObjectType type, 
                              LPCONFIG_OBJECT * ppObject)
{
    int rc = 0;

    return rc;
}

int parserDeviceObjectFind(LPCONFIG_DEVICE pDevice, enum displayObjectType type,
                           int object_id, LPCONFIG_OBJECT * ppObject)
{
    int rc = 0;

    return rc;
}


void parserDeviceObjectUse(enum displayObjectType type, int id, char * pszName)
{
}

void parserDeviceEnd(void)
{
    LP_PARSER_CONTEXT pContext = parserGetContext();

    configDisplaySetDevice(pContext->pDisplay, pContext->pDevice);

    pContext->pDevice = NULL;
}


/* ----------------- Display entity ----------------- */

void parserDisplayBegin(const char * pszDisplay)
{
    configDisplayCreate(&(parserGetContext()->pDisplay), pszDisplay);

}

void parserDisplayEnd()
{
    LP_PARSER_CONTEXT pContext = parserGetContext();

    configFileAddDisplay(pContext->pFile, pContext->pDisplay);

    pContext->pDisplay = NULL;
}


/*
void parserDisplaySetDevice(const char * pszDevice)
{
    LP_PARSER_CONTEXT pContext = parserGetContext();

    config

}
*/



/* ----------------- Module entity ----------------- */

void parserModuleBegin(char * pszModule)
{
    configModuleCreate(&(parserGetContext()->pModule), pszModule);
}

void parserModuleSetLibrary(char * pszLibrary)
{
    configModuleSetLibrary(parserGetContext()->pModule, pszLibrary);
}

void parserModuleControlObject(enum displayObjectType object_type, int object_id)
{
    configModuleControlObject(parserGetContext()->pModule, object_type, object_id);
}

void parserModuleReportObject(enum displayObjectType object_type, int object_id, 
                              const char * report_as)
{
    configModuleReportObject(parserGetContext()->pModule, object_type, object_id, report_as);
}

void parserModuleVirtualObject(int module_id, int device_id, const char * pszObject)
{
    configModuleVirtualObject(parserGetContext()->pModule, module_id, device_id, pszObject);
}

void parserModuleEnd()
{
    LP_PARSER_CONTEXT pContext = parserGetContext();

    configDisplayAddModule(pContext->pDisplay, pContext->pModule);

    pContext->pModule = NULL;
}



/* ----------------- Screen entity ----------------- */



void parserScreenBegin(char * pszScreen)
{
    configScreenCreate(&(parserGetContext()->pScreen), pszScreen);
}


void parserScreenEnd(void)
{
    LP_PARSER_CONTEXT pContext = parserGetContext();

    configDisplayAddScreen(pContext->pDisplay, pContext->pScreen);

    pContext->pScreen = NULL;
}


void parserScreenShowModule(char * pszModule, int size_x, int size_y, int pos_x, int pos_y)
{
    configScreenShowModule(parserGetContext()->pScreen, pszModule, size_x, size_y, pos_x, pos_y);
}


