/*
 * Copyright Ian Burnett 2005, 2006.
 *
 * This file is part of Ian's Interactive LCD controller (IILC).
 * 
 * IILC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * IILC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with IILC; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#if defined(_WIN32)
#   include "pthread_w32.h"
#else
#   include <pthread.h>
#   include <errno.h>
#endif

#include "lcd.h"
#include "config.h"
#include "trace.h"
#include "vector.h"
#include "screen.h"
#include "module.h"
#include "display.h"
#include "parserImpl.h"


#define IILC_MODULE MODULE_TYPE_APP

#if defined(_WIN32)
#   define DEFAULT_CONFIG_FILE      "iilc.conf"
#else
#   define DEFAULT_CONFIG_FILE      "/etc/iilc.conf"
#endif


static int processConfigFile(LPCONFIG_FILE pFile, 
                             LPVECTOR pVector);

static int processConfigDisplay(LPCONFIG_DISPLAY pDisplay, 
                                LPDISPLAY * ppDisplay);


typedef struct _tag_module_map
{
    const char * pszModuleName;

    LPMODULE pModule;
}
MODULE_MAP;

int main(int argc, char ** argv)
{
    /* One configuration file per application instance */
    LPCONFIG_FILE pFile = NULL;

    /* A vector which will contain all the displays we configure */
    LPVECTOR pDisplayVector = NULL;
    
    /* Handy loop variables */
    int i = 0, nCount = 0;

    /* Generic return code */
    int rc = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_main);

    if (argc > 1) {
        printf("%s\n", argv[1]);
        g_bSev0Enabled = 1;
        g_bSev1Enabled = 1;
    }

    /* Create a vector to hold all the displays setup by the config file */
    vectorCreate(&pDisplayVector);

    /*
     * Processes the config file. This step performs most of the syntax checking and
     * very little of the contextual checking on the config file.
     */
    rc = parserParse(&pFile, "../conf/lcd.conf");
    if (rc > 0) { goto end_of_function; }

    /*
     * Verify the data structure is valid. This step does no syntax checking, but ensures
     * that things like cross-references within the file are valid and that mistakes such
     * as duplicated entries are caught. This second step is required if we want an LALR
     * parser, but wish to allow the user to define things in any order they desire.
     */
    rc = configFileValidate(pFile);
    if (rc > 0) { goto end_of_function; }

    /*
     * Performs the setup stage which takes the memory model and converts it into running
     * code. At this point, everything is assumed to be correct. There is the small issue
     * where things cannot be determined until we connect to the display (such as number of
     * temperature sensors connected), but that can be dealt with as and when necessary.
     */
    rc = processConfigFile(pFile, pDisplayVector);
    if (rc > 0) { goto end_of_function; }


    /* Start each of the display instances */
    vectorGetSize(pDisplayVector, &nCount);
    for (i = 0; (i < nCount) && (rc == 0); i++) {
        LPDISPLAY pDisplay = NULL;
        rc = vectorGetElementAt(pDisplayVector, i, (void **) &pDisplay);
        if (!rc) {
            rc = displayStart(pDisplay);
        }
    }

    if (rc) {
        goto end_of_function;
    }

    /* Wait around for a bit (just demo code at the minute) */
    SLEEP(10000);

    /* Stop each of the display instances */
    for (i = 0; (i < nCount) && (rc == 0); i++) {

        LPDISPLAY pDisplay = NULL;

        /* Get each display */
        rc = vectorGetElementAt(pDisplayVector, i, (void **) &pDisplay);

        /* Remove from the vector */
        vectorRemove(pDisplayVector, pDisplay);

        /* Stop it... */
        if (!rc) {
            rc = displayStop(pDisplay);
        }

        /* ... and dispose. */
        if (!rc) {
            rc = displayDispose(&pDisplay);
        }
    }

    if (rc) {
        goto end_of_function;
    }

    /* Dispose of the vector */
    vectorDestroy(&pDisplayVector);

end_of_function:

    /* Delete the config file instance */
    if (pFile != NULL) {
        configFileDispose(&pFile);
    }

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}



/****************************************************************************/
/*                                                                          */
/* Functions to convert the memory model into something useful              */
/*                                                                          */
/****************************************************************************/


/**
  * Takes a config file object and creates a set of LPDISPLAY objects and inserts
  * them into the supplied vector. When this function returns, the LPDISPLAY objects
  * will be fully configured ready to call displayStart().
  */
int processConfigFile(LPCONFIG_FILE pConfigFile, LPVECTOR pVector)
{
    int rc = 0;
    int i = 0, nCount = 0;

    IILC_TRACE_ENTRY(IILC_MODULE, T_processConfigFile);
    traceDebug(IILC_TRACE_CONTEXT, "pFile %p, pVector %p", pConfigFile, pVector);

    /* Number of displays in config file */
    rc = configFileGetDisplayCount(pConfigFile, &nCount);
    if (rc) { goto end_of_function; }

    /* Check we have something configured */
    if (nCount == 0) {
        rc = ERR_INVALID_CONFIG;
        traceError(IILC_TRACE_CONTEXT, "No displays configured for application. Exiting");
        goto end_of_function;
    }

    /* Initialise each display */
    for (i = 0; (i < nCount) && (rc == 0); i++) {

        LPCONFIG_DISPLAY pConfigDisplay = NULL;
        LPDISPLAY        pDisplay       = NULL;

        /* Get the display configuration information out of the model */
        rc = configFileGetDisplay(pConfigFile, i, &pConfigDisplay);
        if (!rc) {
            rc = processConfigDisplay(pConfigDisplay, &pDisplay);
        }

        /* Add this display to the vector */
        vectorAdd(pVector, pDisplay);
    }

end_of_function:

    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


/**
  * Takes a LPCONFIG_DISPLAY object and converts it into a fully-configured
  * LPDISPLAY object.
  */
int processConfigDisplay(LPCONFIG_DISPLAY pConfigDisplay, LPDISPLAY * ppDisplay)
{
    int rc     = 0;
    int i = 0, j = 0;
    int nCount = 0;
    int nModuleCount = 0;

    LPCONFIG_DEVICE pConfigDevice = NULL;

    LPDISPLAY pDisplay = NULL;
    MODULE_MAP * pModuleMap = NULL;

    /* Configuration information obtained from the device data */
    char * pszClass = NULL;
    char * pszPort  = NULL;

    IILC_TRACE_ENTRY(IILC_MODULE, T_processConfigDisplay);
    traceDebug(IILC_TRACE_CONTEXT, 
        "pConfigDisplay %p, ppDisplay %p", pConfigDisplay, ppDisplay);

    /* ----- Device configuration ----- */

    /* Get the configuration of the device to use */
    configDisplayGetDevice(pConfigDisplay, &pConfigDevice);

    /* Extract the necessary information */
    configDeviceGetClass(pConfigDevice, &pszClass);
    configDeviceGetPort(pConfigDevice, &pszPort);

    /* Create the display object */
    displayCreate(pszClass, pszPort, &pDisplay);


    /* ----- Module configuration ----- */

    /* How many modules are involved here? */
    configDisplayGetModuleCount(pConfigDisplay, &nModuleCount);

    /* Alloc some space for this */
    pModuleMap = (MODULE_MAP *) calloc(nCount, sizeof(MODULE_MAP));

    for (i = 0; (i < nModuleCount) && (rc == 0); i++) {

        LPCONFIG_MODULE pConfigModule = NULL;
        const char * pszModuleName = NULL;
        LPMODULE pModule = NULL;
        int nReports = 0;

        /* Get the specific module */
        configDisplayGetModule(pConfigDisplay, i, &pConfigModule);

        /* Get the module configuration information */
        configModuleGetLibrary(pConfigModule, &pszModuleName);

        /* Load the module */
        moduleInit(pszModuleName, &pModule);

        /* Store information away for later use */
        pModuleMap[i].pszModuleName = pszModuleName;
        pModuleMap[i].pModule       = pModule;

        /* --- Setup the entities this module receives reports for --- */

        /* Get the number of objects this module wishes to receive reports for */
        configModuleGetReportCount(pConfigModule, &nReports);
        for (j = 0; (j < nReports) && (rc == 0); j++) {

            LPCONFIG_OBJECT pConfigObject = NULL;
            enum displayObjectType type;
            unsigned int nId;
            char * pszReportName = NULL;

            /* Get the configuration of this object */
            configModuleGetReportObject(pConfigModule, j, &pConfigObject);

            /* Get the configuration information */
            configObjectGetId(pConfigObject, &nId);
            configObjectGetName(pConfigObject, &pszReportName);

            /* Decide what sort of action this should take */
            configObjectGetType(pConfigObject, &type);
            switch (type) {

            case OBJECT_TYPE_FAN:
                displayAddFanReportListener(pDisplay, nId, pModule, pszReportName);
                break;

            case OBJECT_TYPE_TEMP:
                displayAddTempReportListener(pDisplay, nId, pModule, pszReportName);
                break;

            case OBJECT_TYPE_LED: /* LEDs don't get reported */
            default:
                break;
            }

        }


        /* --- Setup the virtual entities that this module provides --- */


        /* --- Setup the controlling links --- */

        //displaySetLEDController(pDisplay, nId, pModule);

    }


    /* ----- Screen configuration ----- */


    /* How many screens are involved here? */
    configDisplayGetScreenCount(pConfigDisplay, &nCount);
    for (i = 0; (i < nCount) && (rc == 0); i++) {

        LPCONFIG_SCREEN pConfigScreen = NULL;
        LPSCREEN        pScreen       = NULL;
        int nWindowCount;

        /* Create a new screen object to manipulate */
        screenCreate(&pScreen, 20, 4);

        /* Get the next screen information from the config object */
        configDisplayGetScreen(pConfigDisplay, i, &pConfigScreen);

        /* How many windows exist in this screen? */
        configScreenGetWindowCount(pConfigScreen, &nWindowCount);
        for (j = 0; (j < nWindowCount) && (rc == 0); j++) {

            LPCONFIG_WINDOW pConfigWindow = NULL;
            LPMODULE        pModule       = NULL;
            int x, y, width, height;
            const char * pszModuleName = NULL;
            int m = 0;

            /* Get the next window */
            configScreenGetWindow(pConfigScreen, j, &pConfigWindow);

            /* Get the various bits of information about it */
            configWindowGetLocation(pConfigWindow, &x, &y);
            configWindowGetSize(pConfigWindow, &width, &height);
            
            /* Get the module used to control this window */
            configWindowGetModuleName(pConfigWindow, &pszModuleName);

            /* Lookup the actual module we've mapped to this config object */
            for (m = 0; m < nModuleCount; m++) {
                if (strcmp(pModuleMap[m].pszModuleName, pszModuleName) == 0) {
                    pModule = pModuleMap[m].pModule;
                    break;
                }
            }

            /* Found a match? */
            if (pModule != NULL) {

                /* Add to the main screen */
                screenAddWindow(pScreen, x, y, width, height, pModule);
            }
            else {
                printf("Module %s not found!\n", pszModuleName);
            }
        }

        /* Add the fully-configured screen to the display */
        displayScreenAdd(pDisplay, pScreen);
    }


    /* ----- Final tidy-up ----- */

    /* Return to the caller */
    *ppDisplay = pDisplay;


    IILC_TRACE_EXIT_RC(rc);
    return rc;
}


